//
// gANTlab module for Hertzian dipole excitation
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "module_header.h"

#include "natconst.h"

#include "module_data.h"
#include "rotmat.h"

using namespace std;



extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }



double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }



const int RHS_para_no = 7;
const char *RHS_para[ RHS_para_no ] =
{
   "x / m",
   "y / m",
   "z / m",
   "theta / °",
   "phi / °",
   "abs(Il) / Am",
   "arg(Il) / °"
};

extern "C" void RHS_get_parameter( const MainData *MD, void *data_,
                                   string **param, string **value,
                                   int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = RHS_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;

   sprintf(sb,"%g",data->x );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->y );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->z );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->theta * 180.0 / M_PI );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->phi * 180.0 / M_PI );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", abs(data->Il) );
   (*param)[id] = RHS_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", arg(data->Il) );
   (*param)[id] = RHS_para[id];  
   (*value)[id] = sb;
   id++;

}


extern "C" void RHS_set_parameter( const MainData *MD, void *data_,
                                   const string *param, const string *value,
                                   int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;
   double absIl, argIl;

   absIl = abs(data->Il);
   argIl = arg(data->Il);

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<RHS_para_no; j++ ) if ( RHS_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->x = atof( value[i].c_str() ); break; }
         case 1: { data->y = atof( value[i].c_str() ); break; }
         case 2: { data->z = atof( value[i].c_str() ); break; }
         case 3: { data->theta = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 4: { data->phi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 5: { absIl = atof( value[i].c_str() ); break; }
         case 6: { argIl = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         default: throw( "unknown parameter" );
      }
   }

   data->Il = complex<double>( absIl * cos(argIl), absIl * sin(argIl) );

   double rot[3][3];
   rotmat( rot, 1.0, 0.0, 0.0, 0.0, data->theta, data->phi );
   data->ux[0] = rot[0][0];
   data->ux[1] = rot[1][0];
   data->ux[2] = rot[2][0];

   data->uy[0] = rot[0][1];
   data->uy[1] = rot[1][1];
   data->uy[2] = rot[2][1];

   data->uz[0] = rot[0][2];
   data->uz[1] = rot[1][2];
   data->uz[2] = rot[2][2];
}


#ifdef HAVE_LIBGL
extern "C" void RHS_add_to_OGLlist( const MainData *MD, void *data_, GLuint gl_list )
{
   ModuleData *data = (ModuleData*)data_;

   const int Hno = 8;

   const GLfloat clr_E[4] = { 0.90, 0.00, 0.00, 1.0 };
   const GLfloat clr_H[4] = { 0.25, 0.25, 1.00, 1.0 };

   double R  = 0.5*HYPOT( MD->size_x, MD->size_y, MD->size_z );

// E
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_E );

   glBegin(GL_LINES);

   glNormal3f( data->uz[0], data->uz[1], data->uz[2] );

   glVertex3d( data->x-0.10*R*data->uz[0],
               data->y-0.10*R*data->uz[1],
               data->z-0.10*R*data->uz[2] );
   glVertex3d( data->x+0.10*R*data->uz[0],
               data->y+0.10*R*data->uz[1],
               data->z+0.10*R*data->uz[2] );

   glEnd();

   glBegin(GL_LINE_LOOP);

   glVertex3d( data->x-0.10*R*data->ux[0],
               data->y-0.10*R*data->ux[1],
               data->z-0.10*R*data->ux[2] );
   glVertex3d( data->x+0.10*R*data->uz[0],
               data->y+0.10*R*data->uz[1],
               data->z+0.10*R*data->uz[2] );
   glVertex3d( data->x+0.10*R*data->ux[0],
               data->y+0.10*R*data->ux[1],
               data->z+0.10*R*data->ux[2] );

   glEnd();

   glBegin(GL_LINE_LOOP);

   glVertex3d( data->x-0.10*R*data->uy[0],  
               data->y-0.10*R*data->uy[1],  
               data->z-0.10*R*data->uy[2] );
   glVertex3d( data->x+0.10*R*data->uz[0],  
               data->y+0.10*R*data->uz[1],  
               data->z+0.10*R*data->uz[2] );
   glVertex3d( data->x+0.10*R*data->uy[0],  
               data->y+0.10*R*data->uy[1],  
               data->z+0.10*R*data->uy[2] );

   glEnd();

// H
   double w, p[3];

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_H );

   glBegin(GL_LINE_LOOP);

   glNormal3f( data->uz[0], data->uz[1], data->uz[2] );

   for ( int i=0; i<Hno; i++ )
   {
      w = i * 2.0*M_PI / Hno;

      p[0] = data->ux[0] * cos(w) + data->uy[0] * sin(w);
      p[1] = data->ux[1] * cos(w) + data->uy[1] * sin(w);
      p[2] = data->ux[2] * cos(w) + data->uy[2] * sin(w);

      glVertex3d( data->x + 0.10*R*p[0],
                  data->y + 0.10*R*p[1],
                  data->z + 0.10*R*p[2]  );
   }

   glEnd();
}
#endif
