//
// gANTlab module for Hertzian dipole excitation (E_inc)
// (c) 28.6.2007 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "module_header.h"

#include "natconst.h"

#include "module_data.h"
#include "rotmat.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "Hertzian Dipole Module (E_inc)";



extern double HYPOT( double x, double y, double z );



extern "C" void RHS_Einc( const MainData *MD, void *data_, double x, double y, double z,
                          complex<double> E[3] )
{
   ModuleData *data = (ModuleData*)data_;

   double rv[3];
   rv[0] = x-data->x;
   rv[1] = y-data->y;
   rv[2] = z-data->z;

   double R  = HYPOT( rv[0], rv[1], rv[2] );
   double kR = MD->k0*R;

   double costh = ( data->uz[0]*rv[0] + data->uz[1]*rv[1] + data->uz[2]*rv[2] ) / R;
   double sinth = sqrt( 1.0 - costh*costh );

   double phi = atan2( data->uy[0]*rv[0] + data->uy[1]*rv[1] + data->uy[2]*rv[2],
                       data->ux[0]*rv[0] + data->ux[1]*rv[1] + data->ux[2]*rv[2]  );
   double cosphi = cos(phi);
   double sinphi = sin(phi);

   complex<double> expo = exp( complex<double>( 0.0, -kR ) ) * data->Il;

   complex<double> Er = ZF0*costh * expo
                        * ( 1.0 - complex<double>( 0.0, 1.0/kR ) )
                        / (2.0*M_PI*R*R);

   complex<double> Eth = complex<double>( 0.0, ZF0*MD->k0*sinth ) * expo
                         * ( 1.0 - complex<double>( 0.0, 1.0/kR ) - 1.0/(kR*kR) )
                         / (4.0*M_PI*R);

   complex<double> Ex = cosphi*( sinth*Er + costh*Eth );
   complex<double> Ey = sinphi*( sinth*Er + costh*Eth );
   complex<double> Ez =          costh*Er - sinth*Eth;


   E[0] = data->ux[0]*Ex + data->uy[0]*Ey + data->uz[0]*Ez;
   E[1] = data->ux[1]*Ex + data->uy[1]*Ey + data->uz[1]*Ez;
   E[2] = data->ux[2]*Ex + data->uy[2]*Ey + data->uz[2]*Ez;
}
