//
// gANTlab module for Hertzian dipole excitation
// (c) 28.6.2007 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "module_header.h"

#include "natconst.h"

#include "module_data.h"
#include "rotmat.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "Hertzian Dipole Module (J)";




extern "C" void matrix_bfkt_bounding_box( const MainData *MD, void *data_, int nbfkt,
                                          int *sno,
                                          double **cpos_x, double **cpos_y, double **cpos_z,
                                          double **size_x, double **size_y, double **size_z  )
{
   ModuleData *data = (ModuleData*)data_;

   if ( nbfkt != 0 ) throw( "one excitation only" );

   *sno = 1;
   
   *cpos_x = new double[*sno];
   *cpos_y = new double[*sno];
   *cpos_z = new double[*sno];

   *size_x = new double[*sno];
   *size_y = new double[*sno];
   *size_z = new double[*sno];

   **cpos_x = data->x;
   **cpos_y = data->y;
   **cpos_z = data->z;

   double dx = ( data->x - ( MD->cpos_x - 0.5*MD->size_x ) );
   dx -= rint(dx/MD->delta_x) * MD->delta_x;
   double dy = ( data->y - ( MD->cpos_y - 0.5*MD->size_y ) );
   dy -= rint(dy/MD->delta_y) * MD->delta_y;
   double dz = ( data->z - ( MD->cpos_z - 0.5*MD->size_z ) );
   dz -= rint(dz/MD->delta_z) * MD->delta_z;

   if ( fabs(dx) < 0.01*MD->delta_x + DBL_EPSILON )
      **size_x = DBL_EPSILON;
   else
      **size_x = 8.0*MD->delta_x;

   if ( fabs(dy) < 0.01*MD->delta_y + DBL_EPSILON )
      **size_y = DBL_EPSILON;
   else
      **size_y = 8.0*MD->delta_y;

   if ( fabs(dz) < 0.01*MD->delta_z + DBL_EPSILON )
      **size_z = DBL_EPSILON;
   else
      **size_z = 8.0*MD->delta_z;

}


extern "C" void matrix_bfkt_moment( const MainData *MD, void *data_,
                                    int nbfkt, int sno,
                                    double cpos_x, double cpos_y, double cpos_z,
                                    double len_x, double len_y, double len_z,
                                    int nx, int ny, int nz,
                                    double *mx, double *my, double *mz, double *md )
{
   ModuleData *data = (ModuleData*)data_;

   if ( nbfkt != 0 ) throw( "one excitation only" );

   double dx = ( data->x - cpos_x ) / len_x;
   double dy = ( data->y - cpos_y ) / len_y;
   double dz = ( data->z - cpos_z ) / len_z;

   double powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

   *mx = powpow * data->uz[0];
   *my = powpow * data->uz[1];
   *mz = powpow * data->uz[2];

   *md = 0.0;
}



extern "C" void RHS_bfkt_inc( const MainData *MD, void *data_,
                              complex<double> **ampl, int *no )
{
   ModuleData *data = (ModuleData*)data_;

   *no = 1;
   *ampl = new complex<double>[*no];

   (*ampl)[0] = data->Il;
}
