//
// gANTlab test module II --- E probe using field grid
// (c) 28.6.2007 Ingo D. Rullhusen
//

#include <stdio.h>

#include <iostream>

#include "module_header.h"

#include "module_data2.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "Test Module II";



extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }







const int extract_para_no = 3;
const char *extract_para[ extract_para_no ] =
{
   "P x / m",
   "P y / m",
   "P z / m"
};


extern "C" void extract_get_parameter( const MainData *MD, void *data_,
                                       string **param, string **value,
                                       int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = extract_para_no;
   *param = new string[*no];
   *value = new string[*no];   

   int id = 0;

   sprintf(sb,"%g",data->apx );
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->apy );
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->apz );
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

}

extern "C" void extract_set_parameter( const MainData *MD, void *data_,
                                       const string *param, const string *value,
                                       int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<extract_para_no; j++ ) if ( extract_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->apx = atof( value[i].c_str() ); break; }
         case 1: { data->apy = atof( value[i].c_str() ); break; }
         case 2: { data->apz = atof( value[i].c_str() ); break; }
         default: throw( "unknown parameter" );
      }
   }

}


static double SQU( double x ){ return(x*x); }

extern "C" void extract_run( const MainData *MD, void *data_, bool dummy,
                             complex<double> *JV, int no, const FIELD *f,
                             string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;

   const complex<double> cNULL(0.0,0.0);
   const complex<double> cJ(0.0,1.0);

   int fidx;
   double phi, sinphi, cosphi, costh, sinth, xs, ys, zs, R;
   complex<double> E[3], Er, Eth, Exs, Eys, Ezs, expo;

   E[0] = E[1] = E[2] = cNULL;

   for ( int i3=0; i3<f->noz; i3++ )
   {
      for ( int i2=0; i2<f->noy; i2++ )
      {
         for ( int i1=0; i1<f->nox; i1++ )
         {
            fidx = i1 + f->nox * i2 + f->nox*f->noy * i3;


// x directed
   xs = data->apy - ( i2*f->delta_y + f->lpos_y );
   ys = data->apz - ( i3*f->delta_z + f->lpos_z );
   zs = data->apx - ( i1*f->delta_x + f->lpos_x );

   phi = atan2( ys, xs ); 
   sinphi = sin(phi);
   cosphi = cos(phi);

   R = sqrt( xs*xs + ys*ys + zs*zs );
   costh = zs / R;   
   sinth = sqrt( 1.0 - costh*costh );

   expo = exp( -cJ * MD->k0 * R );

   Er = costh * 2.0*data->prefac_E * expo * ( 1.0 + 1.0 / (cJ*MD->k0*R) ) / (R*R);
   Eth = sinth * data->prefac_E * expo * cJ*MD->k0 * ( 1.0 + 1.0 / (cJ*MD->k0*R) - 1.0 / SQU(MD->k0*R) ) / R;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   E[0] += Ezs * f->J[ fidx ].x;
   E[1] += Exs * f->J[ fidx ].x;
   E[2] += Eys * f->J[ fidx ].x;


// y directed
   xs = - ( data->apx - ( i1*f->delta_x + f->lpos_x ) );
   ys = data->apz - ( i3*f->delta_z + f->lpos_z );
   zs = data->apy - ( i2*f->delta_y + f->lpos_y );

   phi = atan2( ys, xs );
   sinphi = sin(phi);
   cosphi = cos(phi);

   R = sqrt( xs*xs + ys*ys + zs*zs );
   costh = zs / R;
   sinth = sqrt( 1.0 - costh*costh );

   expo = exp( -cJ * MD->k0 * R );

   Er = costh * 2.0*data->prefac_E * expo * ( 1.0 + 1.0 / (cJ*MD->k0*R) ) / (R*R);
   Eth = sinth * data->prefac_E * expo * cJ*MD->k0 * ( 1.0 + 1.0 / (cJ*MD->k0*R) - 1.0 / SQU(MD->k0*R) ) / R;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   E[0] += -Exs * f->J[ fidx ].y;
   E[1] +=  Ezs * f->J[ fidx ].y;
   E[2] +=  Eys * f->J[ fidx ].y;

// z directed
   xs = data->apx - ( i1*f->delta_x + f->lpos_x );
   ys = data->apy - ( i2*f->delta_y + f->lpos_y );
   zs = data->apz - ( i3*f->delta_z + f->lpos_z );

   phi = atan2( ys, xs );
   sinphi = sin(phi);
   cosphi = cos(phi);

   R = sqrt( xs*xs + ys*ys + zs*zs );
   costh = zs / R;
   sinth = sqrt( 1.0 - costh*costh );

   expo = exp( -cJ * MD->k0 * R );

   Er = costh * 2.0*data->prefac_E * expo * ( 1.0 + 1.0 / (cJ*MD->k0*R) ) / (R*R);
   Eth = sinth * data->prefac_E * expo * cJ*MD->k0 * ( 1.0 + 1.0 / (cJ*MD->k0*R) - 1.0 / SQU(MD->k0*R) ) / R;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   E[0] += Exs * f->J[ fidx ].z;
   E[1] += Eys * f->J[ fidx ].z;
   E[2] += Ezs * f->J[ fidx ].z;


         }
      }
   }


   *nomsg = 3;
   *omsg = new string[*nomsg];
   char buff[50];
   sprintf( buff, "( %e; %e )", E[0].real(), E[0].imag() );
   (*omsg)[0] = "E_x " + (string)buff;
   sprintf( buff, "( %e; %e )", E[1].real(), E[1].imag() );
   (*omsg)[1] = "E_y " + (string)buff;
   sprintf( buff, "( %e; %e )", E[2].real(), E[2].imag() );
   (*omsg)[2] = "E_z " + (string)buff;



}
