//
// AIM data class
// (c) 27.3.2007 Ingo D. Rullhusen
//


#include <iostream>

#include <math.h>
#include <float.h>

#include "global_vars.h"


#ifndef NO_SMALL_CAPS

#ifdef NOF77_
#define DGETRF dgetrf
#define DGETRS dgetrs
#else
#define DGETRF dgetrf_
#define DGETRS dgetrs_
#endif

#endif

extern "C" void DGETRF( int*, int*, double*, int*, int*, int* );
extern "C" void DGETRS( char*, int*, int*, double*, int*, int*, double*, int*, int* );


AIMData::AIMData( MainDataX MD, const ExternalModule *M, int nobfkt )
{
   int subno;
   double *cx, *cy, *cz, *size_x, *size_y, *size_z;

   M->matrix_bfkt_bounding_box( &MD, nobfkt,
                                &subno,
                                &cx, &cy, &cz,
                                &size_x, &size_y, &size_z );  // allocate

   double Dx = ( MD.AIMnox > 1 ? MD.size_x / (MD.AIMnox - 1) : 0.0 ),
          Dy = ( MD.AIMnoy > 1 ? MD.size_y / (MD.AIMnoy - 1) : 0.0 ),
          Dz = ( MD.AIMnoz > 1 ? MD.size_z / (MD.AIMnoz - 1) : 0.0 );

   double ZPx = MD.cpos_x - 0.5*MD.size_x,
          ZPy = MD.cpos_y - 0.5*MD.size_y,
          ZPz = MD.cpos_z - 0.5*MD.size_z;

   int minx, maxx, miny, maxy, minz, maxz,
       *loff_x, *loff_y, *loff_z, *lno_x, *lno_y, *lno_z;

   loff_x = new int[subno];
   loff_y = new int[subno];
   loff_z = new int[subno];

   lno_x = new int[subno];
   lno_y = new int[subno];
   lno_z = new int[subno];

   x_off = MD.AIMnox;
   y_off = MD.AIMnoy;
   z_off = MD.AIMnoz;

   nox = 0;
   noy = 0;
   noz = 0;

   for ( int s=0; s<subno; ++s )
   {
      if ( Dx > DBL_EPSILON )
      {
         minx = (int)floor( ( cx[s] - 0.5*size_x[s] - ZPx ) / Dx + 0.1 );
         maxx = (int)ceil(  ( cx[s] + 0.5*size_x[s] - ZPx ) / Dx - 0.1 );
      }
      else
      {
         minx = maxx = 0;
      }

      if ( Dy > DBL_EPSILON )
      {
         miny = (int)floor( ( cy[s] - 0.5*size_y[s] - ZPy ) / Dy + 0.1 );
         maxy = (int)ceil(  ( cy[s] + 0.5*size_y[s] - ZPy ) / Dy - 0.1 );
      }
      else
      {
         miny = maxy = 0;
      }

      if ( Dz > DBL_EPSILON )
      {
         minz = (int)floor( ( cz[s] - 0.5*size_z[s] - ZPz ) / Dz + 0.1 );
         maxz = (int)ceil(  ( cz[s] + 0.5*size_z[s] - ZPz ) / Dz - 0.1 );
      }
      else
      {
         minz = maxz = 0;
      }

      lno_x[s] = maxx - minx + 1;
      lno_y[s] = maxy - miny + 1;
      lno_z[s] = maxz - minz + 1;

      loff_x[s] = minx;
      loff_y[s] = miny;
      loff_z[s] = minz;

      x_off = ( x_off < minx ? x_off : minx );
      y_off = ( y_off < miny ? y_off : miny );
      z_off = ( z_off < minz ? z_off : minz );

      nox = ( nox > maxx ? nox : maxx );
      noy = ( noy > maxy ? noy : maxy );
      noz = ( noz > maxz ? noz : maxz );
   }

   nox = nox - x_off + 1;
   noy = noy - y_off + 1;
   noz = noz - z_off + 1;

   for ( int s=0; s<subno; ++s )
   {
      loff_x[s] -= x_off;
      loff_y[s] -= y_off;
      loff_z[s] -= z_off;
   }

   if ( x_off < 0 || y_off < 0 || z_off < 0 ||
        x_off >= MD.AIMnox || y_off >= MD.AIMnoy || z_off >= MD.AIMnoz )
      throw( "index problem in aimdata" );


   int id, noxyz, id2;
   double *mx, *my, *mz, *md, len_x, len_y, len_z;
   double *ex, *ey, *ez, *ediv;

   noxyz = nox * noy * noz;

   lambda_x = new AIMreal[ noxyz ];
   lambda_y = new AIMreal[ noxyz ];
   lambda_z = new AIMreal[ noxyz ];
   lambda_div = new AIMreal[ noxyz ];

   for ( int i=0; i<noxyz; ++i )
   {
      lambda_x[i] = 0.0;
      lambda_y[i] = 0.0;
      lambda_z[i] = 0.0;
      lambda_div[i] = 0.0;
   }

   for ( int s=0; s<subno; ++s )
   {
      len_x = ( size_x[s] > DBL_EPSILON ? size_x[s] : 1.0 );
      len_y = ( size_y[s] > DBL_EPSILON ? size_y[s] : 1.0 );
      len_z = ( size_z[s] > DBL_EPSILON ? size_z[s] : 1.0 );

      noxyz = lno_x[s]*lno_y[s]*lno_z[s];
      mx = new double[ 4 * noxyz ];
      my = mx + noxyz;
      mz = my + noxyz;
      md = mz + noxyz;

      ex = new double[noxyz];
      ey = new double[noxyz];
      ez = new double[noxyz];
      ediv = new double[noxyz];

      for ( int iz=0; iz<lno_z[s]; iz++ )
      {
         for ( int iy=0; iy<lno_y[s]; iy++ )
         {
            for ( int ix=0; ix<lno_x[s]; ix++ )
            {
               id = ix + iy*lno_x[s] + iz*lno_x[s]*lno_y[s];

               M->matrix_bfkt_moment( &MD, nobfkt, s,
                                      cx[s], cy[s], cz[s], len_x, len_y, len_z,
                                      ix, iy, iz,
                                      &mx[id], &my[id], &mz[id], &md[id] );

               ex[id] = mx[id];
               ey[id] = my[id];
               ez[id] = mz[id];
               ediv[id] = md[id];
            }
         }
      }


      int wdim, wid;
      double *w, dx, dy, dz, wdimd;
      wdim = noxyz * noxyz;
      wdimd = (double)lno_x[s]*(double)lno_y[s]*(double)lno_z[s]
              * (double)lno_x[s]*(double)lno_y[s]*(double)lno_z[s];
      if ( wdim < 0.99*wdimd )
         throw( "range problem in aimdata" );

      w = new double[ wdim ];


      for ( int px=0; px<lno_x[s]; ++px )
      {
         for ( int py=0; py<lno_y[s]; ++py )
         {
            for ( int pz=0; pz<lno_z[s]; ++pz )
            {
               id = px + py*lno_x[s] + pz*lno_x[s]*lno_y[s];

               dx = ( (px+loff_x[s]+x_off)*Dx+ZPx - cx[s] ) / len_x;
               dy = ( (py+loff_y[s]+y_off)*Dy+ZPy - cy[s] ) / len_y;
               dz = ( (pz+loff_z[s]+z_off)*Dz+ZPz - cz[s] ) / len_z;

               for ( int ix=0; ix<lno_x[s]; ++ix )
               {
                  for ( int iy=0; iy<lno_y[s]; ++iy )
                  {
                     for ( int iz=0; iz<lno_z[s]; ++iz )
                     {
                        wid = id*noxyz + (ix+iy*lno_x[s]+iz*lno_x[s]*lno_y[s]);

                        w[ wid ] = pow( dx, ix ) * pow( dy, iy ) * pow( dz, iz );
                     }
                  }
               }

            }
         }
      }


      char TRANS;
      int INFO, LDB, *IPIV, LDA, NRHS, N;

      N = noxyz;
      LDA = N;
      LDB = N;
      NRHS = 4;
      IPIV = new int[N];

      TRANS = 'N';
      INFO = 0;    
      DGETRF( &N, &N, w, &LDA, IPIV, &INFO );
      if ( INFO )
         throw( "problem with DGETRF" );

      DGETRS( &TRANS, &N, &NRHS, w, &LDA, IPIV, mx, &LDB, &INFO );
      if ( INFO )
         throw( "problem with DGETRS" );

      delete [] IPIV;
      delete [] w;

      for ( int iz=0; iz<lno_z[s]; iz++ )
      {
         for ( int iy=0; iy<lno_y[s]; iy++ )
         {
            for ( int ix=0; ix<lno_x[s]; ix++ )
            {
               id = ix + iy*lno_x[s] + iz*lno_x[s]*lno_y[s];
               id2 = ( ix+loff_x[s] )
                     + ( iy+loff_y[s] ) * nox
                       + ( iz+loff_z[s] ) * nox*noy;

               lambda_x[id2] += mx[id];
               lambda_y[id2] += my[id];
               lambda_z[id2] += mz[id];
               lambda_div[id2] += md[id];
            }
         }
      }


// error check
      double ex_max, ey_max, ez_max, ediv_max, ex_max_err, ey_max_err, ez_max_err, ediv_max_err,
             Jx, Jy, Jz, divJ, powpow;

      ex_max = 0.0;
      ey_max = 0.0;
      ez_max = 0.0;
      ediv_max = 0.0;
      ex_max_err = 0.0;
      ey_max_err = 0.0;
      ez_max_err = 0.0;
      ediv_max_err = 0.0;
      for ( int iz=0; iz<lno_z[s]; ++iz )
      {
         for ( int iy=0; iy<lno_y[s]; ++iy )
         {
            for ( int ix=0; ix<lno_x[s]; ++ix )
            {
               Jx = 0.0;
               Jy = 0.0;
               Jz = 0.0;
               divJ = 0.0;

               for ( int pz=0; pz<lno_z[s]; ++pz )
               {
                  for ( int py=0; py<lno_y[s]; ++py )
                  {
                     for ( int px=0; px<lno_x[s]; ++px )
                     {
                        id = px + py*lno_x[s] + pz*lno_x[s]*lno_y[s];

                        dx = ( (px+loff_x[s]+x_off)*Dx+ZPx - cx[s] ) / len_x;
                        dy = ( (py+loff_y[s]+y_off)*Dy+ZPy - cy[s] ) / len_y;
                        dz = ( (pz+loff_z[s]+z_off)*Dz+ZPz - cz[s] ) / len_z;

                        powpow = pow( dx, ix ) * pow( dy, iy ) * pow( dz, iz );

                        Jx += mx[id] * powpow;
                        Jy += my[id] * powpow;
                        Jz += mz[id] * powpow;
                      divJ += md[id] * powpow;
                     }
                  }
               }


               id = ix + iy*lno_x[s] + iz*lno_x[s]*lno_y[s];

               if ( fabs( Jx - ex[id] ) > ex_max_err )
               {
                  ex_max_err = fabs( Jx - ex[id] );
                  ex_max = fabs( ex[id] );
               }
               if ( fabs( Jy - ey[id] ) > ey_max_err )
               {
                  ey_max_err = fabs( Jy - ey[id] );
                  ey_max = fabs( ey[id] );
               }
               if ( fabs( Jz - ez[id] ) > ez_max_err )
               {
                  ez_max_err = fabs( Jz - ez[id] );
                  ez_max = fabs( ez[id] );
               }
               if ( fabs( divJ - ediv[id] ) > ediv_max_err )
               {
                  ediv_max_err = fabs( divJ - ediv[id] );
                  ediv_max = fabs( ediv[id] );
               }

            }
         }
      }   
      if ( ex_max < MD.epsilon ) ex_max = 1.0/MD.epsilon;
      if ( ey_max < MD.epsilon ) ey_max = 1.0/MD.epsilon;
      if ( ez_max < MD.epsilon ) ez_max = 1.0/MD.epsilon;
      if ( ediv_max < MD.epsilon ) ediv_max = 1.0/MD.epsilon;

      if ( ex_max_err/ex_max > MD.epsilon )
      {
         Xlog << endl;
         Xlog << "WARNING in AIMData(): Error overflow X" << endl
              << "                      absolute = "
              << ex_max_err << endl
              << "                      relative = "
              << ex_max_err/ex_max << endl
              << "                      base     = "
              << ex_max << endl;
      }
      if ( ey_max_err/ey_max > MD.epsilon )
      {
         Xlog << endl;
         Xlog << "WARNING in AIMData(): Error overflow Y" << endl
              << "                      absolute = "
              << ey_max_err << endl
              << "                      relative = "
              << ey_max_err/ey_max << endl
              << "                      base     = "
              << ey_max << endl;
      }
      if ( ez_max_err/ez_max > MD.epsilon )
      {
         Xlog << endl;
         Xlog << "WARNING in AIMData(): Error overflow Z" << endl
              << "                      absolute = "
              << ez_max_err << endl
              << "                      relative = "
              << ez_max_err/ez_max << endl
              << "                      base     = "
              << ez_max << endl;
      }
      if ( ediv_max_err/ediv_max > MD.epsilon )
      {
         Xlog << endl;
         Xlog << "WARNING in AIMData(): Error overflow DIV" << endl
              << "                      absolute = "
              << ediv_max_err << endl
              << "                      relative = "
              << ediv_max_err/ediv_max << endl
              << "                      base     = "
              << ediv_max << endl;
      }

      delete [] mx;

      delete [] ex;
      delete [] ey;
      delete [] ez;
      delete [] ediv;

   } // end:s

   delete [] cx;
   delete [] cy;
   delete [] cz;
   delete [] size_x;
   delete [] size_y;
   delete [] size_z;
}

AIMData::AIMData( const AIMData &AD2 )
{
   nox = AD2.nox;
   noy = AD2.noy;
   noz = AD2.noz;

   lambda_x = new AIMreal[nox*noy*noz];
   lambda_y = new AIMreal[nox*noy*noz];
   lambda_z = new AIMreal[nox*noy*noz];
   lambda_div = new AIMreal[nox*noy*noz];

   for ( int i=0; i<nox*noy*noz; i++ )
   {
      lambda_x[i] = AD2.lambda_x[i];
      lambda_y[i] = AD2.lambda_y[i];
      lambda_z[i] = AD2.lambda_z[i];
      lambda_div[i] = AD2.lambda_div[i];
   }

   x_off = AD2.x_off;
   y_off = AD2.y_off;
   z_off = AD2.z_off;
}

AIMData::AIMData()
{
   nox = noy = noz = 0;
   lambda_x = lambda_y = lambda_z = lambda_div = NULL;
   x_off = y_off = z_off = 0;
}


AIMData::~AIMData()
{
   delete [] lambda_x;
   delete [] lambda_y;
   delete [] lambda_z;
   delete [] lambda_div;
}


void AIMData::write_binary( fstream &file ) const
{
   file.write( (char*) this, sizeof(*this) );

   int noxyz = nox*noy*noz;
   file.write( (char*)lambda_x, sizeof(*lambda_x)*noxyz );
   file.write( (char*)lambda_y, sizeof(*lambda_y)*noxyz );
   file.write( (char*)lambda_z, sizeof(*lambda_z)*noxyz );
   file.write( (char*)lambda_div, sizeof(*lambda_div)*noxyz );
}


AIMData::AIMData( fstream &file )
{
   file.read( (char*) this, sizeof(*this) );

   int noxyz = nox*noy*noz;

   lambda_x = new AIMreal[noxyz];
   lambda_y = new AIMreal[noxyz];
   lambda_z = new AIMreal[noxyz];
   lambda_div = new AIMreal[noxyz];

   file.read( (char*)lambda_x, sizeof(*lambda_x)*noxyz );
   file.read( (char*)lambda_y, sizeof(*lambda_y)*noxyz );
   file.read( (char*)lambda_z, sizeof(*lambda_z)*noxyz );
   file.read( (char*)lambda_div, sizeof(*lambda_div)*noxyz ); 
}


int AIMData::index( int ix, int iy, int iz ) const
{
#ifdef BOUNDARY_CHECK
   if ( ix < 0 || iy < 0 || iz < 0 ||
        ix >= nox || iy >= noy || iz >= noz )
   {
      cerr << "Error in AIMData::index()!" << endl;
      exit(1);
   }
#endif

   return( ix + nox*iy + nox*noy*iz );
}


AIMData& AIMData::operator =( const AIMData AD2 )
{
   delete [] lambda_x;
   delete [] lambda_y;
   delete [] lambda_z;
   delete [] lambda_div;

   nox = AD2.nox;
   noy = AD2.noy;
   noz = AD2.noz;

   x_off = AD2.x_off;
   y_off = AD2.y_off;
   z_off = AD2.z_off;

   lambda_x = new AIMreal[nox*noy*noz];
   lambda_y = new AIMreal[nox*noy*noz];
   lambda_z = new AIMreal[nox*noy*noz];
   lambda_div = new AIMreal[nox*noy*noz];

   for ( int i=0; i<nox*noy*noz; i++ )
   {
      lambda_x[i] = AD2.lambda_x[i];
      lambda_y[i] = AD2.lambda_y[i];
      lambda_z[i] = AD2.lambda_z[i];
      lambda_div[i] = AD2.lambda_div[i];
   }

   return(*this);
}
