//
// Class for AIM solver
// (c) 5.5.2006 Ingo D. Rullhusen
//

#ifndef SOLVER_H
#define SOLVER_H

#include <complex>
#include <fstream>
#include <vector>

#include "aimfield.h"

#include "farfield.h"

using namespace std;


class AIMsolver
{
   public:
      int veno;
      complex<double> *RHS;
      complex<double> *J;

      int itno;
      double epsilon;

      int no_of_inc_modules;    // modules for excitation
      int *inc_modno;           // global mod no
      int *start_idx_inc_mod;   // & dummy at end
      vector<AIMData> ADinc;
      vector< complex<double> > AMPLinc;

      FARFIELD *ff;

      AIMsolver();
      ~AIMsolver();

      void read_binary( fstream &file );
      void write_binary( fstream &file ) const; 

      void setup();
      void solve( double &epsilon, int &max_itno );
      void clear();

      void extract();

   private:
      void precon( complex<double> *in, complex<double> *out );
      void matvecmult( AIMfield *af, complex<double> *x, complex<double> *ax );

};


#endif
