//
// auxiliary stuff
// (c) 18.4.2007 Ingo D. Rullhusen
//


#include <iostream>

#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "auxiliary.h"


void goto_next_line( fstream &file )   // until next item
{
   char cb = ' ';

   while ( file.good() && (cb!='\n') ){ file.get(cb); }

   cb = file.peek();
   while( file.good() )
   {
      while ( file.good() && (isspace(cb) || cb == '\n') )
      { file.get(cb); cb = file.peek();}

      if ( cb == '#' )
      {
         while ( file.good() && (cb!='\n') ){ file.get(cb); }
         cb = file.peek();
      }
      else break;
   }

   if ( file.bad() ) throw( "file corrupt" );
}


string get_section_label( fstream &file, bool skip )
{
   string sbuff;
   fstream::pos_type pos;

   pos = file.tellg();

   while ( file.good() )
   {
      file >> sbuff;

      if ( sbuff.length() < 1 || sbuff.at(0) == '#' )
      {
         goto_next_line( file );
      }
      else if ( sbuff.at(0) == '[' &&
                sbuff.at( sbuff.length()-1 ) == ']' )
      {
         if ( skip )
         {
            goto_next_line( file );
         }
         else file.seekg( pos );

         return( sbuff.substr( 1, sbuff.size()-2 ) );
      }
   }

   return("");
}


string get_quoted_string( fstream &file )
{
   char cb = ' ';
   string sb = "";

   while ( (file.good()) && (cb!='"') ){ file.get(cb); }
   file.get(cb);

   while ( (file.good()) && (cb!='"') ){ sb += cb; file.get(cb); }

   return(sb);
}


string checksum( string fname )
{
   unsigned long csum;
   const int buffsize = 30;
   char sbuff[buffsize];
   struct stat filestatus;

   if ( stat(fname.c_str(), &filestatus) < 0 ) return("!!!error!!!");

   csum = 0;
   csum += filestatus.st_size;
   csum += filestatus.st_mtime;

   snprintf(sbuff,buffsize,"%lx",csum);
   sbuff[buffsize-1] = '\0';

   return(sbuff);
}


void normal( double tp1[3], double tp2[3], double tp3[3], double nv[3] )
{
   double a[3], b[3];

   a[0] = tp2[0] - tp1[0];
   a[1] = tp2[1] - tp1[1];
   a[2] = tp2[2] - tp1[2];

   b[0] = tp3[0] - tp1[0];
   b[1] = tp3[1] - tp1[1];
   b[2] = tp3[2] - tp1[2];

   nv[0] = a[1]*b[2] - a[2]*b[1];
   nv[1] = a[2]*b[0] - a[0]*b[2];
   nv[2] = a[0]*b[1] - a[1]*b[0];

   double AR2 = HYPOT( nv[0], nv[1], nv[2] );
   nv[0] /= AR2;
   nv[1] /= AR2;
   nv[2] /= AR2;
}


void umkugel( double no, double *cp_x, double *cp_y, double *cp_z,
              double *s_x, double *s_y, double *s_z,
              double &R, double CP[3] )
{
   double minx, miny, minz, maxx, maxy, maxz, dt;

   minx = cp_x[0] - 0.5 * s_x[0];
   miny = cp_y[0] - 0.5 * s_y[0];
   minz = cp_z[0] - 0.5 * s_z[0];

   maxx = cp_x[0] + 0.5 * s_x[0];
   maxy = cp_y[0] + 0.5 * s_y[0];
   maxz = cp_z[0] + 0.5 * s_z[0];

   for ( int i=1; i<no; ++i )
   {
      minx = ( (dt = cp_x[i] - 0.5 * s_x[i]) < minx ? dt : minx );
      miny = ( (dt = cp_y[i] - 0.5 * s_y[i]) < miny ? dt : miny );
      minz = ( (dt = cp_z[i] - 0.5 * s_z[i]) < minz ? dt : minz );

      maxx = ( (dt = cp_x[i] + 0.5 * s_x[i]) > maxx ? dt : maxx );
      maxy = ( (dt = cp_y[i] + 0.5 * s_y[i]) > maxy ? dt : maxy );
      maxz = ( (dt = cp_z[i] + 0.5 * s_z[i]) > maxz ? dt : maxz );
   }

   CP[0] = 0.5 * ( maxx + minx );
   CP[1] = 0.5 * ( maxy + miny );
   CP[2] = 0.5 * ( maxz + minz );

   R = 0.5 * HYPOT( maxx-minx, maxy-miny, maxz-minz );
}

