//
// auxiliary qt stuff
// (c) 28.3.2007 Ingo D. Rullhusen
//


#include <iostream>


#include <unistd.h>

#include <qthread.h>
#include <qprogressdialog.h>
#include <qfiledialog.h>
#include <qimage.h>

#include "auxiliary_qt.h"
#include "global_vars.h"



double FreqUnitFactor( QString str )
{
   if      ( str == (QString)"mHz" ) return( 1e+3 );
   else if ( str == (QString)"Hz"  ) return( 1e0 );
   else if ( str == (QString)"kHz" ) return( 1e-3 );
   else if ( str == (QString)"MHz" ) return( 1e-6 );
   else if ( str == (QString)"GHz" ) return( 1e-9 );
   else if ( str == (QString)"THz" ) return( 1e-12 );

   throw( "GUI error");
}


double LengthUnitFactor( QString str )
{
   if      ( str == (QString)"km" ) return( 1e-3 );
   else if ( str == (QString)"m"  ) return( 1e0 );
   else if ( str == (QString)"dm" ) return( 1e+1 );
   else if ( str == (QString)"cm" ) return( 1e+2 );
   else if ( str == (QString)"mm" ) return( 1e+3 );
   else if ( str == (QString)"m" ) return( 1e+6 );

   throw( "GUI error");
}




bool thread_buzy_run( QString *name, QThread *T )
{
   QProgressDialog PW( *name,
                       QProgressDialog::tr("Cancel"),
                       0,
                       ptrMW,
                       QProgressDialog::tr("busy"),
                       true );

   unsigned int ui = 0;

   T->start();
   while ( T->running() )
   {
      PW.setProgress(ui);
      ptrAPP->processEvents();
      ui += 10;
      usleep(20000);

      if ( PW.wasCanceled() )
      {
         T->terminate();
         T->wait();
         return(false);
      }
   }

   PW.cancel();

   return(true);
}


// GEO


class thread_GEO : public QThread
{
   public:
      bool ok;
      ExternalModule *em;
      int nomsg;
      string *omsg;

      thread_GEO( ExternalModule *em_ )
      : ok(true), em(em_), nomsg(0), omsg(NULL) {};
      ~thread_GEO(){ delete [] omsg; };
      virtual void run();
};

void thread_GEO::run()
{
   try
   {
     em->matrix_geo_setup( &MD, &omsg, &nomsg );
   }
   catch ( const char *str )
   {
      ok = false;
      Xlog << "geo thread error: " << str << endl;
   }
   exit();
}


bool thread_of_geo_setup( ExternalModule *em )
{
   QString str = QProgressDialog::tr("Set-up geometry of %1").arg(em->name());
   thread_GEO TG( em );

   bool ok = thread_buzy_run( &str, &TG );

   if ( TG.nomsg > 0 )
   {
      Xlog << "output of \"" << em->name() << "\" >>>" << endl;
   }
   for ( int m=0; m<TG.nomsg; m++ )
   {
      Xlog << TG.omsg[m] << endl;
   }
   if ( TG.nomsg > 0 )
   {
      Xlog << "<<< end of output" << endl;
   }

   return( ok && TG.ok );
}


// ELE


class thread_ELE : public QThread
{
   public:
      bool ok;
      ExternalModule *em;
      int nomsg;
      string *omsg;

      thread_ELE( ExternalModule *em_ )
      : ok(true), em(em_), nomsg(0), omsg(NULL) {};
      ~thread_ELE(){ delete [] omsg; };
      virtual void run();
};

void thread_ELE::run()
{
   em->matrix_ele_setup( &MD, &omsg, &nomsg );
   exit();
}

bool thread_of_ele_setup( ExternalModule *em )
{
   QString str = QProgressDialog::tr("Set-up elements of %1").arg(em->name());
   thread_ELE TE( em );

   bool ok = thread_buzy_run( &str, &TE );

   if ( TE.nomsg > 0 )
   {
      Xlog << "output of \"" << em->name() << "\" >>>" << endl;
   }
   for ( int m=0; m<TE.nomsg; m++ )
   {
      Xlog << TE.omsg[m] << endl;
   }
   if ( TE.nomsg > 0 )
   {
      Xlog << "<<< end of output" << endl;
   }

   return( ok && TE.ok );
}


// RHS


class thread_RHS : public QThread
{
   public:
      bool ok;
      ExternalModule *em;
      int nomsg;
      string *omsg;

      thread_RHS( ExternalModule *em_ )
      : ok(true), em(em_), nomsg(0), omsg(NULL) {};
      ~thread_RHS(){ delete [] omsg; };
      virtual void run();
};

void thread_RHS::run()
{
   em->RHS_E_setup( &MD, &omsg, &nomsg );
   exit();
}

bool thread_of_rhs_setup( ExternalModule *em )
{
   QString str = QProgressDialog::tr("Set-up RHS of %1").arg(em->name());
   thread_RHS TR( em );

   bool ok = thread_buzy_run( &str, &TR );

   if ( TR.nomsg > 0 )
   {
      Xlog << "output of \"" << em->name() << "\" >>>" << endl;
   }
   for ( int m=0; m<TR.nomsg; m++ )
   {
      Xlog << TR.omsg[m] << endl;
   }
   if ( TR.nomsg > 0 )
   {
      Xlog << "<<< end of output" << endl;
   }

   return( ok && TR.ok );
}


// EXTRACT


class thread_extract : public QThread
{
   public:
      bool ok;
      ExternalModule *em;
      FIELD *f;
      int nomsg;
      string *omsg;

      thread_extract( ExternalModule *em_, FIELD *f_ )
      : ok(true), em(em_), f(f_) , nomsg(0), omsg(NULL) {};
      ~thread_extract(){ delete [] omsg; };
      virtual void run();
};

void thread_extract::run()
{
   try
   {
      em->extract_run( &MD,
                       &SOL.J[em->matrix_offset()],
                       em->no_of_bfkt(),
                       f, &omsg, &nomsg );
   }
   catch ( const char *str )
   {
      ok = false;
      Xlog << "extract thread error: " << str << endl;
   }
   exit();
}

bool thread_of_extract_setup( ExternalModule *em, FIELD *f )
{
   QString str = QProgressDialog::tr("Run extract of %1").arg(em->name());
   thread_extract TE( em, f );

   bool ok = thread_buzy_run( &str, &TE );

   if ( TE.nomsg > 0 )
   {
      Xlog << "output of \"" << em->name() << "\" >>>" << endl;
   }
   for ( int m=0; m<TE.nomsg; m++ )
   {
      Xlog << TE.omsg[m] << endl;
   }
   if ( TE.nomsg > 0 )
   {
      Xlog << "<<< end of output" << endl;
   }

   return( ok && TE.ok );
}
