//
// Class for density plot
// (c) 24.4.2007 Ingo D. Rullhusen
//


#include <iostream>

#include <qimage.h>

#include "density.h"

#include "global_vars.h"


Density::Density( QWidget *parent, const char *name )
: Drawing( parent, name )
{
   wml1 = wml2 = NULL;
}


Density::~Density()
{
   clear();
}



void Density::replot( bool profile, bool block )
{
   clear();
   wml1 = wml2 = NULL;

   double psx = width()  / (double)MAT.Z->nofrows;
   double psy = height() / (double)MAT.Z->nofrows;

   QImage dpimg( width(), height(), 8, 3, QImage::IgnoreEndian );
   dpimg.setColor ( 0, BkgColor().rgb() );
   dpimg.setColor ( 1, qRgb ( 255,   0,   0 ) );
   dpimg.setColor ( 2, qRgb ( 240, 200, 200 ) );
   dpimg.fill( 0 );

   for ( int i=0; i<MAT.Z->nofrows; i++ )
   {
      for ( int k=MAT.Z->first_ele[i]; k<MAT.Z->first_ele[i+1]; k++ )
      {

         for ( int ix = (int)rint(MAT.Z->colid[k]*psy);
               ix < (int)rint(MAT.Z->colid[k]*psy) + (int)ceil(psy);
               ix++ )
         {
            for ( int iy = (int)rint(i*psx);
                  iy < (int)rint(i*psx) + (int)ceil(psx);
                  iy++ )
            {
               if ( dpimg.valid(iy,ix) ) dpimg.setPixel ( iy, ix, 2 );
            }
         }

         for ( int ix = (int)rint(MAT.Z->colid[k]*psx);
               ix < (int)rint(MAT.Z->colid[k]*psx) + (int)ceil(psx);
               ix++ )
         {
            for ( int iy = (int)rint(i*psy);
                  iy < (int)rint(i*psy) + (int)ceil(psy);
                  iy++ )
            {
               if ( dpimg.valid(ix,iy) ) dpimg.setPixel ( ix, iy, 1 );
            }
         }

      }
   }

   canvas.setBackgroundPixmap( dpimg );


   QPen pen;
   QBrush brush;
 
   brush.setStyle( Qt::SolidPattern );
   brush.setColor( Qt::red );

   pen.setWidth( 1 );
   pen.setCapStyle( Qt::FlatCap );
   pen.setJoinStyle( Qt::RoundJoin );

// profile
   if ( profile )
   {
      QCanvasLine *pf1 = new QCanvasLine(&canvas), *pf2 = new QCanvasLine(&canvas);

      pen.setStyle( Qt::SolidLine );
      pen.setColor( Qt::blue );

      pf1->setPoints( (int)rint(MD.precon*psx), 0,
                      width()-1, height()-(int)rint(MD.precon*psy) );
      pf2->setPoints( 0, (int)rint(MD.precon*psy),
                      width()-(int)rint(MD.precon*psx), height()-1 );

      pf1->setPen( pen );
      pf2->setPen( pen );

      pf1->setZ(1.0);
      pf2->setZ(1.0);

      pf1->show();
      pf2->show();
   }


// block
   if ( block )
   {
      pen.setStyle( Qt::DashDotLine );
      pen.setColor( Qt::gray );

      for ( int i=1; i<MAT.no_of_modules; ++i )
      {
         QCanvasLine *pf = new QCanvasLine(&canvas);
         pf->setPen( pen );
         pf->setZ(0.0);

         pf->setPoints( 0,         (int)rint(psy*MAT.start_idx_mod[i]),
                        width()-1, (int)rint(psy*MAT.start_idx_mod[i])  );

         pf->show();
      }

      for ( int i=0; i<MAT.no_of_modules; ++i )
      {
         QCanvasText *txt = new QCanvasText(&canvas);
         txt->setColor( Qt::gray );
         txt->setZ(0.0);

         QString str = QString::number( MAT.modno[i] + 1 );
         str += ": " + used_modules[ MAT.modno[i] ].name();

         txt->setText( str );

         QRect bb = txt->boundingRect();

         txt->setY( psy*MAT.start_idx_mod[i] + 0.01*height() );
         if ( MAT.start_idx_mod[i] < MAT.Z->nofrows/2 )
         {
            txt->setX( width() - bb.width() - 0.01*width() );
         }
         else
         {
            txt->setX( 0.01*width() );
         }

         txt->show();
      }
   }

   wm = false;

}


void Density::replot_mark( int w )
{
   double psx = width()  / (double)MAT.Z->nofrows;
   double psy = height() / (double)MAT.Z->nofrows;

   wm = true;

   QPen pen;

   pen.setWidth( 1 );
   pen.setCapStyle( Qt::FlatCap );
   pen.setJoinStyle( Qt::RoundJoin );

   delete wml1;
   delete wml2;

   wml1 = new QCanvasLine(&canvas);
   wml2 = new QCanvasLine(&canvas);

   pen.setStyle( Qt::SolidLine );
   pen.setColor( Qt::green );

   wml1->setPoints( (int)rint(w*psx), 0,
                    width()-1, height()-(int)rint(w*psy) );
   wml2->setPoints( 0, (int)rint(w*psy),
                    width()-(int)rint(w*psx), height()-1 );

   wml1->setPen( pen );
   wml2->setPen( pen );

   wml1->setZ(2.0);
   wml2->setZ(2.0);

   wml1->show();
   wml2->show();
}


void Density::wheelEvent( QWheelEvent* e )
{
   if ( wm )
   {
      emit got_wheel_event( e->delta() );
      e->accept();
   }
}

