/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you want to add, delete, or rename functions or slots, use
** Qt Designer to update this file, preserving your code.
**
** You should not define a constructor or destructor in this file.
** Instead, write your code in functions called init() and destroy().
** These will automatically be called by the form's constructor and
** destructor.
*****************************************************************************/


#include "global_vars.h"


void DensityForm::exec_density()
{
   replot();

   exec();
}


void DensityForm::replot()
{
   if ( S.value() < MAT_RDY || MAT.Z->nofrows < 1 ) return;

   density->replot( Profile->isChecked(), Blocks->isChecked() );

   if ( Width->isChecked() )
   {
      int w = WidthValue->text().toInt();
      if ( w < 0 ) w = 0;
      if ( w > MAT.Z->nofrows ) w = MAT.Z->nofrows;
      WidthValue->setText( sysLocale.toString(w) );
      density->replot_mark( w );
   }
}



void DensityForm::adjust_precon( int delta )
{
      int w = sysLocale.toInt(WidthValue->text());

      if ( delta < 0 )
         w = 5*((int)floor( 0.01*abs(MD.precon) + 5 + w )/5);
      else
         w = 5*((int)ceil( w - 0.01*abs(MD.precon) - 1 )/5);

      if ( w < 0 ) w = 0;
      if ( w > MAT.Z->nofrows ) w = MAT.Z->nofrows;
      WidthValue->setText( sysLocale.toString(w) );
      density->replot_mark( w );
}




void DensityForm::set_precon()
{
   int w = sysLocale.toInt(WidthValue->text());
   if ( w < 0 ) w = 0;
   if ( w > MAT.Z->nofrows ) w = MAT.Z->nofrows;
   WidthValue->setText( sysLocale.toString(w) );

   Xlog << "set precon profile to " << w << endl;

   MD.precon = w;
   S.back_to( AIM_RDY );

   close();
}




QString DensityForm::GetSaveDensityName()
{
   QString fn;
   QFileDialog fd( MD.PrjPath,
                   tr("All files (*)"),
                   this,
                   tr("save image dialog"),
                   true );
   fd.setMode( QFileDialog::AnyFile );
   fd.setCaption(tr("Save matrix density"));

   QStringList list = QImage::outputFormatList();
   QStringList::Iterator it = list.begin();
   while( it != list.end() )
   {
      fd.addFilter( tr("Image (*.") + *it + " *." + (*it).lower() + ")" );
      ++it;
   }
   fd.setSelectedFilter(tr("All files (*)"));

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}




void DensityForm::export_image()
{

   fstream file;
   QPixmap PIX = QPixmap::grabWidget(density);
   QString fn = GetSaveDensityName();

   if ( fn != "" )
   {
      QString ext = fn.section( '.', -1 );

      if ( PIX.save( fn, ext.upper(), 100 ) )
      {
         Xlog << "matrix density exported as \"" << fn << "\"." << endl;
      }
      else
      {
         Xlog << "cannot save \"" << fn << "\", wrong extension?" << endl;
      }
   }
}



void DensityForm::get_mouse_position( QMouseEvent *e )
{
   if ( ! Width->isChecked() ) return;

   if ( e->state() == Qt::RightButton )
   {
      QPoint MP = e->pos();

      int w = (int)fabs( rint(MAT.Z->nofrows * ( MP.x() / (double)density->width()
                                                  - MP.y() / (double)density->height() )
                              ) );

      if ( w < 0 ) w = 0;
      if ( w > MAT.Z->nofrows ) w = MAT.Z->nofrows;
      WidthValue->setText( sysLocale.toString(w) );
      density->replot_mark( w );
   }
}

