//
// Base class for 2D drawings
// (c) 24.5.2006 Ingo D. Rullhusen
//

#include <iostream>
#include "drawing.h"

using namespace std;

Drawing::Drawing( QWidget *parent, const char *name )
: QCanvasView( parent, name )
{
    canvas.setAdvancePeriod (30);
    setCanvas(&canvas);
    
    setResizePolicy(QScrollView::AutoOneFit);
    setVScrollBarMode(QScrollView::AlwaysOff);
    setHScrollBarMode(QScrollView::AlwaysOff);
}

Drawing::~Drawing()
{
    clear();
}

void Drawing::setBkgColor( QColor c )
{
    canvas.setBackgroundColor(c);
}

QColor Drawing::BkgColor() const
{
    return( canvas.backgroundColor() );
}


void Drawing::viewportResizeEvent( QResizeEvent *re )
{
   canvas.resize(re->size().width(),re->size().height());
   emit resized();
}

void Drawing::clear()
{
    QCanvasItemList list = canvas.allItems();
    QCanvasItemList::Iterator it = list.begin();
    for (; it != list.end(); ++it)
    {
       if ( *it ) delete *it;
    } 
}


void Drawing::mousePressEvent( QMouseEvent* e )
{
   emit mouseButtonPressed( e );
   e->accept(); 
}

void Drawing::mouseMoveEvent( QMouseEvent* e )
{
   emit mouseMoved( e );
   e->accept();
}
