//
// Graph utils
// (c) 30.6.2006 Ingo D. Rullhusen
//

#include <iostream>
#include <algorithm>

#include <qprogressdialog.h>

#include "global_vars.h"
#include "graph.h"


Node::Node()
{
   index = -1;
   neighlist.resize(0);
}


Node::~Node()
{
   neighlist.resize(0);
}


Graph::Graph( int no_of_nodes )
: no_of_nodes(no_of_nodes)
{
   node = new Node[no_of_nodes];
   for ( int i=0; i<no_of_nodes; i++ ) node[i].number = i;
}


Graph::~Graph()
{
   delete [] node;
   no_of_nodes = 0;
}


class case_lesser
{
   public:
      bool operator()( const Node *nn1, const Node *nn2 ) const
      {
         return( nn1->neighlist.size() < nn2->neighlist.size() );
      }
};


void Graph::level_structure( int &len, Node **last_min, Node *root )
{
   vector<Node*> level, next_level;

   for ( int n=0; n<no_of_nodes; n++ )
   {
      if ( node[n].index < 0 )
         node[n].used = false;
      else
         node[n].used = true;
   }

   len = 0;
   next_level.resize(0);
   next_level.push_back( root );
   root->used = true;

   while( next_level.size() > 0 )
   {
      len++;
      level.swap( next_level );
      next_level.resize(0);

      for ( vector<Node*>::iterator an=level.begin();
            an != level.end();
            an++ )
      {
         for ( vector<Node*>::iterator nn=(*an)->neighlist.begin();
               nn != (*an)->neighlist.end();
               nn++ )
         {
            if ( ! (*nn)->used )
            {
               next_level.push_back( *nn );
               (*nn)->used = true;
            }
         }
      }
   }

   sort( level.begin(), level.end(), case_lesser() );

   *last_min = *level.begin();

}


void Graph::insert_edge( int no1, int no2 )
{
#ifdef BOUNDARY_CHECK
   if ( no1 < 0 || no1 >= no_of_nodes ||
        no2 < 0 || no2 >= no_of_nodes    )
   {
      cerr << "Error in Graph::insert_edge(): indices out of range!" << endl;
      exit(1);
   }
#endif

   node[no1].neighlist.push_back(&(node[no2]));
   node[no2].neighlist.push_back(&(node[no1]));
}



void Graph::reverse_Cuthill_McKee()
{
   int label, l, lr;
   int *node_no;
   Node *start, *next, *nextnext;

   node_no = new int[no_of_nodes];
   for ( int i=0; i<no_of_nodes; i++ ) node_no[i] = -1;


// find root
   label = no_of_nodes - 1;
   start = &(node[label]);

/***** NOP for block matrix
   level_structure( lr,     &next, start );
   level_structure( l,  &nextnext,  next );

   while ( l > lr )
   {
      lr = l;
      next = nextnext;

      level_structure( l, &nextnext, next );
   }
   start = next;
*****/

   start->index = label;
   node_no[label] = start->number;
   label--;


// RCM
   unsigned int ui;
   QProgressDialog *PW;
   PW = new QProgressDialog( QProgressDialog::tr("Reverse Cuthill McKee"),
                             QProgressDialog::tr("Cancel"),
                             no_of_nodes-1,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   ui = 0;
   PW->setProgress(ui);

   for ( int n=no_of_nodes-1; n>0; n-- )
   {
      ui++;
      PW->setProgress(ui);
      if ( PW->wasCanceled() )
      {
         throw( "RCM calculation canceled" );
      }
      ptrAPP->processEvents();

      if ( node_no[n] < 0 )
      {

// find new root
         for ( int i=no_of_nodes-1; i>-1; i-- )
         {
            if ( node[i].index < 0 )
            {
               start = &(node[i]);
               break;
            }
         }

/***** NOP for block matrix
         level_structure( lr,     &next, start );
         level_structure( l,  &nextnext,  next );

         while ( l > lr )
         {
            lr = l;
            next = nextnext;

            level_structure( l, &nextnext, next );
         }
         start = next;
*****/

         start->index = label;
         node_no[label] = start->number;
         label--;

      }
      else start = &( node[ node_no[n] ] );

      sort( start->neighlist.begin(),
            start->neighlist.end(),
            case_lesser() );

      for ( vector<Node*>::iterator nn=start->neighlist.begin();
            nn != start->neighlist.end();
            nn++ )
      {
         if ( (*nn)->index < 0 )
         {
            (*nn)->index = label;
            node_no[label] = (*nn)->number;
            label--;
         }
      }

   }

   PW->cancel();
   delete PW;

#ifdef BOUNDARY_CHECK
   if ( label != -1 )
   {
      cerr << "Error in Graph::reverse_Cuthill_McKee(): internal!" << endl;
      exit(1);
   }
#endif

   delete [] node_no;
}
