//
// log stream class
// (c) 12.3.2007 Ingo D. Rullhusen
//

#include <sstream>

#include "global_vars.h"

#include "MyEvents.h"


void LOGstream::open( const char *filename )
{
   logfile.open( filename, ios::out | ios::trunc );
   if ( ! logfile.is_open() ) throw( "cannot open log file" );
}

LOGstream::~LOGstream()
{
   logfile.close();
}


LOGstream& LOGstream::operator<<( const QString str )
{
   QString *pstr = new QString(str);
   logfile << (const char *)sysCodec->fromUnicode(*pstr) << flush;
   QCustomEvent *e = new QCustomEvent(APPEND_LOG_EVENT);
   e->setData( (void*)pstr );
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent

   pstr = NULL;
   return(*this);
}

LOGstream& LOGstream::operator<<( const double d )
{
   QString *pstr = new QString(QString::number(d,'g',10));
   logfile << *pstr << flush;
   QCustomEvent *e = new QCustomEvent(APPEND_LOG_EVENT);
   e->setData( pstr );
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent

   pstr = NULL;
   return(*this);
}

LOGstream& LOGstream::operator<<( const complex<double> c )
{
   QString *pstr = new QString("(");
   *pstr = *pstr + QString::number(c.real(),'g',10) + ";"
                 + QString::number(c.imag(),'g',10) + ")";
   logfile << *pstr << flush;
   QCustomEvent *e = new QCustomEvent(APPEND_LOG_EVENT);
   e->setData( pstr );
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent

   pstr = NULL;
   return(*this);
}


LOGstream& LOGstream::operator<<( ostream&(*rhs)(ostream&) )
{
   stringbuf sb;
   ostream os( &sb );

   rhs(os);

   QString *pstr = new QString(sb.str());
   logfile << (const char *)sysCodec->fromUnicode(*pstr) << flush;
   QCustomEvent *e = new QCustomEvent(APPEND_LOG_EVENT);
   e->setData( pstr );
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent

   pstr = NULL;
   return *this;
}

