//
// main() of gANTlab
// (c) 26.4.2007 Ingo D. Rullhusen
//


#include <stdlib.h>

#include <qapplication.h>
#include <qtranslator.h>
#include <qdir.h> 

#include <iostream>

using namespace std;


#define DEFINE_GLOBAL_VARS
#include "global_vars.h"


int main( int argc, char ** argv )
{
   int retval;

   std::set_new_handler(gantlab_new_error_handler);

   qt_path = getenv("QTDIR");
   if ( qt_path == "" )
   {
      cerr << "QTDIR not set!" << endl;
      return(-2);
   }
   root_path = getenv("GANTLAB_ROOT_DIR");
   if ( root_path == "" )
   {
      QDir rootdir( QDir::home() );
      rootdir.setFilter(QDir::Dirs);
      rootdir.setNameFilter("gANTlab-*.*.*");
      rootdir.setSorting(QDir::Reversed);
      if ( rootdir.count() > 0 )
         root_path = rootdir.absPath() + "/" + rootdir[0];
      else
      {
         cerr << "GANTLAB_ROOT_DIR not set!" << endl;
         return(-2);
      }
   }

   QString lang = getenv("LANG");
   if ( lang.find( "UTF-8", 0, false ) > -1 )
   {
      sysCodec = QTextCodec::codecForName("utf8");
   }
   else
   {
      sysCodec = QTextCodec::codecForName("ISO8859-15");
   }
   if ( sysCodec == NULL )
   {
      cerr << "sysCodec not found!" << endl;
      return(-3);
   }

   if ( argc == 2 )
      prjfile = argv[1];
   else
      prjfile = "";

   try
   {
      ptrAPP = NULL;
      ptrAPP = new QApplication( argc, argv );

      QTextCodec::setCodecForCStrings( QTextCodec::codecForName("utf8") );
      QTextCodec::setCodecForTr( QTextCodec::codecForName("utf8") );

      QTranslator translator_base( 0 );
      translator_base.load( QString("qt_") + QTextCodec::locale(), qt_path + "/translations" );
      ptrAPP->installTranslator( &translator_base );

      QTranslator translator_own( 0 );
      translator_own.load( QString("gANTlab_") + QTextCodec::locale(), root_path );
      ptrAPP->installTranslator( &translator_own );

      ptrMW = NULL;
      ptrMW = new mainform;

      Xlog.open("messages.log");

      Xlog << endl;
      Xlog << "gANTlab (version " << VERSION << ")" << endl;
      Xlog << "GNU antenna laboratory" << endl;
      Xlog << "(c) 2006/7 Ingo D. Rullhusen <d01c@uni-bremen.de>" << endl;
      Xlog << endl;

      try
      { if ( prjfile != "" ) ptrMW->load_prj_file( prjfile ); }
      catch( ... )
      { prjfile = ""; }

      ptrMW->show();
      ptrAPP->connect( ptrAPP, SIGNAL( lastWindowClosed() ), ptrAPP, SLOT( quit() ) );
      retval = ptrAPP->exec();
   }
   catch ( const char* error )
   {
      cerr << "MAIN: " << error << "!" << endl;
      retval = -1;
   }

   delete ptrMW;
   delete ptrAPP;

   fftw_cleanup();
    
   return(retval);
}
