//
// main data class
// (c) 30.4.2007 Ingo D. Rullhusen
//


#include <iostream>



#include "natconst.h"
#include "maindatax.h"
#include "auxiliary.h"

using namespace std;

void MainDataX::reset()
{
   freq = wavelength = omega = k0 = 0.0;
   AIMnox = AIMnoy = AIMnoz = 1;
   cpos_x = cpos_y = cpos_z = 0.0;
   size_x = size_y = size_z = 0.0;
   delta_x = delta_y = delta_z = 0.0;
   near_limit = 0.5;

   AIMcm = ACM_SPCDOM;
   precon = -1;
   max_itno = 500;
   epsilon = 1e-5;

   PrjPath = ".";
   PrjTmpPath = "./tmp";
   ModulePath = "./modules";

   coeff = false;
   coeff_numbering = 0;
   coeff_vec = 0;
   coeff_from = 0;
   coeff_to = 0;

   near_field = false;
   nfp_x = nfp_y = nfp_z = 0.0;

   far_field = false;
   pad_x = pad_y = pad_z = 0;
   itpl_deg = 10;
   no_theta = 50;
   no_phi = 100;
   radiated_pwr = false;
}


void MainDataX::write( fstream &file )
{
   file << "[MainDataX]" << endl;

   file << freq << "   #freq"       << endl;

   file << AIMnox << " " << AIMnoy << " " << AIMnoz << "   #AIM grid" << endl;
   file << cpos_x << " " << cpos_y << " " << cpos_z << "   #center position" << endl;
   file << size_x << " " << size_y << " " << size_z << "   #size" << endl;
   file << delta_x << " " << delta_y << " " << delta_z << "   #grid const" << endl;
   file << near_limit << "   #near limit" << endl;

   file << AIMcm << "   #AIM correction mode" << endl
        << precon << "   #precon profile" << endl
        << max_itno << "   #max_itno" << endl
        << epsilon << "   #epsilon" << endl;

   file << "\"" << PrjPath << "\"" << "   #project path" << endl;
   file << "\"" << PrjTmpPath << "\"" << "   #project tmp path" << endl;
   file << "\"" << ModulePath << "\"" << "   #module path" << endl;

   file << coeff << "   #coefficients" << endl;
   file << coeff_numbering << " " << coeff_vec << "   #numbering vec" << endl;
   file << coeff_from << " " << coeff_to << "   #from to" << endl;

   file << near_field << "   #near field" << endl;
   file << nfp_x << " " << nfp_y << " " << nfp_z << "   #near field pos" << endl;

   file << far_field << "   #far-field" << endl;
   file << pad_x << " " << pad_y << " " << pad_z << "   #zero padding" << endl;
   file << itpl_deg << "   #itpl. degree" << endl;
   file << no_theta << " " << no_phi << "   #no. of theta and phi" << endl;
   file << radiated_pwr << "   #calc. rad. power" << endl;
}


void MainDataX::read( fstream &file )
{
   string sbuff = "";
   while ( file.good() && sbuff != "MainDataX" )
   { sbuff = get_section_label( file, true ); }

   if ( sbuff != "MainDataX" ) throw( "cannot find MainDataX entry" );

   file >> sbuff;
   freq = atof( sbuff.c_str() );

   wavelength = C0/freq;
   omega = 2.0 * M_PI * freq;
   k0 = 2.0 * M_PI / wavelength;
   goto_next_line( file );

   file >> AIMnox >> AIMnoy >> AIMnoz;
   goto_next_line( file );

   file >> cpos_x >> cpos_y >> cpos_z;
   goto_next_line( file );

   file >> size_x >> size_y >> size_z;
   goto_next_line( file );

   file >> delta_x >> delta_y >> delta_z;
   goto_next_line( file );

   file >> near_limit;
   goto_next_line( file );

   int acm;
   file >> acm;
   AIMcm = (AIMcorrMode)acm;
   goto_next_line( file );
   file >> precon;
   goto_next_line( file );
   file >> max_itno;
   goto_next_line( file );
   file >> epsilon;
   goto_next_line( file );

   PrjPath = get_quoted_string( file );
   goto_next_line( file );
   PrjTmpPath = get_quoted_string( file );
   goto_next_line( file );
   ModulePath = get_quoted_string( file );
   goto_next_line( file );

   file >> coeff;
   goto_next_line( file );
   file >> coeff_numbering >> coeff_vec;
   goto_next_line( file );
   file >> coeff_from >> coeff_to;
   goto_next_line( file );

   file >> near_field;
   goto_next_line( file );
   file >> nfp_x >> nfp_y >> nfp_z;
   goto_next_line( file );

   file >> far_field;
   goto_next_line( file );
   file >> pad_x >> pad_y >> pad_z;
   goto_next_line( file );
   file >> itpl_deg;
   goto_next_line( file );
   file >> no_theta >> no_phi;
   goto_next_line( file );
   file >> radiated_pwr;

   if ( file.fail() ) throw("read error in maindata section");
}
