/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you want to add, delete, or rename functions or slots, use
** Qt Designer to update this file, preserving your code.
**
** You should not define a constructor or destructor in this file.
** Instead, write your code in functions called init() and destroy().
** These will automatically be called by the form's constructor and
** destructor.
*****************************************************************************/


#define PRJ_FILE_VERSION -122


#ifdef TERMINAL
namespace mainform
{
   void load_prj_file( QString );
   void fileSave();
}
#endif


void mainform::load_prj_file( QString fn )
{
   fstream pf, tmpf;

   pf.open( sysCodec->fromUnicode(fn), ios::in );

   if ( ! pf.is_open() )
   {
      Xlog << "cannot open \"" << fn << "\"!" << endl;
      throw("file not opened");
   }

   MD.reset();
#ifndef TERMINAL
   getFsettings();
   state_effect();
#endif

   used_modules.clear();
   S.back_to( INITIAL );

   prjfile = fn;

   Xlog << "loading project from \"" << prjfile << "\"..." << endl;

   try
   {
// version
      string sbuff = "";
      while ( pf.good() && sbuff != "version" )
      { sbuff = get_section_label( pf, true ); }

      if ( sbuff != "version" )
         throw( "cannot find version entry" );

      pf >> sbuff;
      if ( atoi( sbuff.c_str() ) != PRJ_FILE_VERSION )
         throw( "wrong prj file version" );

// maindata
      MD.read( pf );

// 2D pattern parameter
      pattern2d.read( pf );

// 3D pattern parameter
      pattern3d.read( pf );

// modules
      while ( get_section_label( pf, false ) == "ExternalModule" )
      {
         ExternalModule ExM( &MD, pf );

         used_modules.push_back(ExM);
      }

// state & tmp files
      S.read( pf );

      Xlog << "project loaded successfully." << endl;
   }
   catch ( const char* error )
   {
#ifndef TERMINAL
      Xlog << "open problem: " << error << "!" << endl;
#else
      throw(error);
#endif
   }

   pf.close();

#ifndef TERMINAL
   UpdateModuleTable();
   used_modules_table->clearSelection();
   getFsettings();

   setCaption("gANTlab: " + prjfile);
#endif
}

void mainform::fileSave()
{
   fstream pf;

#ifndef TERMINAL
   if ( prjfile == "" ) prjfile = GetSavePrjName();
#endif

   pf.open( sysCodec->fromUnicode(prjfile), ios::out );

   if ( ! pf.is_open() )
   {
      Xlog << "cannot open \"" << prjfile << "\"!" << endl;
      return;
   }

   pf.precision(15);
   pf.setf(ios::scientific, ios::floatfield);

   try
   {

// version
   pf << "[version]" << endl;
   pf << PRJ_FILE_VERSION << "   #prj file version" << endl;
   pf << "\"" << VERSION << "\"   #gANTlab version" << endl;

   pf << endl;
// maindata
   MD.write( pf );

   pf << endl;
// 2D pattern parameter
   pattern2d.write( pf );

   pf << endl;
// 3D pattern parameter
   pattern3d.write( pf );

   pf << endl;
// modules
   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      ip->write_parameter( &MD, pf );
   }

   pf << endl;
// state & tmp files
   S.write( pf );

   }
   catch ( const char *str )
   {
      pf.close();
      Xlog << "save error in file \"" << prjfile << "\": " << str << "!" << endl;
      Xlog << "project NOT saved correctly" << endl;
      return;
   }

   pf.close();
   Xlog << "project saved as \"" << prjfile << "\"." << endl;

#ifndef TERMINAL
   setCaption("gANTlab: " + prjfile);
#endif
}



#ifndef TERMINAL
// GUI stuff

#include <iostream>
#include <unistd.h>



void mainform::init()
{
   QString fn = root_path
                + "/doc/html/"
                + sysLocale.name().section('_',0,0)
                + "/gantlab.adp";
   if ( ! QFile::exists(fn) ) fn = root_path + QString("/doc/html/gantlab.adp");

   qac = new QAssistantClient(qt_path+"/bin", this);
   QStringList sl;
   sl << "-profile" << fn;
   qac->setArguments( sl );

   MD.reset();
   getFsettings();
   state_effect();
}

void mainform::destroy()
{
   qac->closeAssistant();
   delete qac;
}



void mainform::fileNew()
{
   prjfile = "";
   setCaption("gANTlab");
   S.back_to( INITIAL );
   used_modules.clear();
   UpdateModuleTable();
   MD.reset();
}

QString mainform::GetOpenPrjName()
{
   QString fn;
   QFileDialog fd( MD.PrjPath,
                   tr("gANTlab Projects (*.prj);;All files (*)"),
                   this,
                   tr("open project dialog"),
                   true );
   fd.setMode( QFileDialog::ExistingFile );
   fd.setCaption(tr("Open gANTlab project"));
   fd.setSelection( prjfile );

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}

void mainform::fileOpen()
{
   QString fn = GetOpenPrjName();
   string sbuff;
   fstream pf, tmpf;

   if ( fn == "" ) return;

   load_prj_file( fn );
}

QString mainform::GetSavePrjName()
{
   QString fn;
   QFileDialog fd( MD.PrjPath,
                   tr("gANTlab Projects (*.prj);;All files (*)"),
                   this,
                   tr("save project dialog"),
                   true );
   fd.setMode( QFileDialog::AnyFile );
   fd.setCaption(tr("Save gANTlab project"));
   fd.setSelection( prjfile );

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}

void mainform::fileSaveAs()
{
   QString fn = GetSavePrjName();

   if ( fn != "" )
   {
      prjfile = fn;
      fileSave();
   }
}



void mainform::openProjectSetup()
{
   setEnabled(false);

   try
   {
      prjsu.getMD();

      if ( prjsu.exec() == QDialog::Accepted )
      {
         reset_state();
         prjsu.setMD();
      }
   }
   catch ( const char *e )
   {
      Xlog << "cannot set-up project data: " << e << "!" << endl;
   }

   getFsettings();
   setEnabled(true);
}



void mainform::UpdateModuleTable()
{
   int addrow;
   QString s_t;

   addrow = used_modules.size() - used_modules_table->numRows();
   for ( ; addrow < 0; addrow++ ) used_modules_table->removeRow(0);
   if ( addrow > 0 ) used_modules_table->insertRows(0,addrow);

   addrow = 0;
   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      used_modules_table->setText( addrow, 0, ip->name() );
      s_t.setNum(ip->version());
      used_modules_table->setText( addrow, 1, s_t );


      if ( ip->has_matrix() )
         used_modules_table->setPixmap( addrow, 2,
                                        ( ip->use_matrix() ? QPixmap::fromMimeSource("haken.png")
                                                           : QPixmap::fromMimeSource("kreuz.png") ) );
//         used_modules_table->setText( addrow, 2, ( ip->use_matrix() ? "on" : "off" ) );
      else
         used_modules_table->setPixmap( addrow, 2, QPixmap::fromMimeSource("balken.png") );
//         used_modules_table->setText( addrow, 2, "---" );


      if ( ip->has_RHS() )
         used_modules_table->setPixmap( addrow, 3,
                                        ( ip->use_RHS() ? QPixmap::fromMimeSource("haken.png")
                                                        : QPixmap::fromMimeSource("kreuz.png") ) );
//         used_modules_table->setText( addrow, 3, ( ip->use_RHS() ? "on" : "off" ) );
      else
         used_modules_table->setPixmap( addrow, 3, QPixmap::fromMimeSource("balken.png") );
//         used_modules_table->setText( addrow, 3, "---" );


      if ( ip->has_extract() )
         used_modules_table->setPixmap( addrow, 4,
                                        ( ip->use_extract() ? QPixmap::fromMimeSource("haken.png")
                                                            : QPixmap::fromMimeSource("kreuz.png") ) );
//         used_modules_table->setText( addrow, 4, ( ip->use_extract() ? "on" : "off" ) );
      else
         used_modules_table->setPixmap( addrow, 4, QPixmap::fromMimeSource("balken.png") );
//         used_modules_table->setText( addrow, 4, "---" );


      addrow++;
   }

   for ( int col=0; col<used_modules_table->numCols(); col++ )
      used_modules_table->adjustColumn(col);

   update();
}



QString mainform::GetLoadModuleName()
{
   QString fn;
   QFileDialog fd( MD.ModulePath,
                   tr("gANTlab Modules (gANTlab*.so);;All files (*)"),
                   this,
                   tr("load module dialog"),
                   true );
   fd.setMode( QFileDialog::ExistingFile );
   fd.setCaption(tr("Load gANTlab module"));

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}



void mainform::UsedModulesRequest( int row, int col, int button, const QPoint &pos )
{

   if ( col > -1 && col < 5 )
   {
      gANTlabPEditor e;
      e.prop_tab->setColumnReadOnly(0,TRUE);

      if ( row < 0 ) return;

      e.modfile = used_modules[row].modfile();

      string *param, *value;
      int no;

      switch ( col )
      {
         case 0:
         case 1:
         {
            QString fn = e.modfile.remove(".so") + "_"
                         + sysLocale.name().section('_',0,0)
                         + ".html";
            if ( ! QFile::exists(fn) ) fn = e.modfile.remove(".so") + ".html";

            qac->showPage(fn);

            return;
         }
         case 2:
         {
            e.setCaption(tr("gANTlab: Edit: module matrix properties"));

            if ( ! used_modules[row].has_matrix() ) return;

            used_modules[row].matrix_get_parameter( &MD, &param, &value, &no );
            e.ModulePartActive->setChecked( used_modules[row].use_matrix() );

            break;
         }
         case 3:
         {
            e.setCaption(tr("gANTlab: Edit: module RHS properties"));
   
            if ( ! used_modules[row].has_RHS() ) return;

            used_modules[row].RHS_get_parameter( &MD, &param, &value, &no );
            e.ModulePartActive->setChecked( used_modules[row].use_RHS() );

            break;
         }
         case 4:
         {
            e.setCaption(tr("gANTlab: Edit: module extract properties"));

            if ( ! used_modules[row].has_extract() ) return;

            used_modules[row].extract_get_parameter( &MD, &param, &value, &no );
            e.ModulePartActive->setChecked( used_modules[row].use_extract() );

            break;
         }
         default:
         {
            throw( "column select error" );
         }
      }

      e.prop_tab->insertRows(0,no);
      for ( int i=0; i<no; i++ )
      {
         e.prop_tab->setText( i, 0, param[i] );
         e.prop_tab->setText( i, 1, value[i] );
      }

      delete [] param;
      delete [] value;

      e.prop_tab->adjustColumn(0);

      setEnabled(false);
      if ( e.exec() == QDialog::Accepted )
      {
         no = e.prop_tab->numRows();

         param = new string[no];
         value = new string[no];
         for ( int i=0; i<no; i++ )
         {
            param[i] = (const char *)e.prop_tab->text( i, 0 );
            value[i] = (const char *)e.prop_tab->text( i, 1 );
         }

         switch ( col )
         {
            case 2:
            {
               used_modules[row].matrix_set_parameter( &MD, param, value, no );
               used_modules[row].use_matrix() = e.ModulePartActive->isChecked();
               reset_state();
               break;
            }
            case 3:
            {
               used_modules[row].RHS_set_parameter( &MD, param, value, no );
               used_modules[row].use_RHS() = e.ModulePartActive->isChecked();
               S.back_to( MAT_RDY );
               break;
            }
            case 4:
            {
               used_modules[row].extract_set_parameter( &MD, param, value, no );
               used_modules[row].use_extract() = e.ModulePartActive->isChecked();
               S.back_to( SOLVED );
               break;
            }
            default:
            {
               setEnabled(true);
               throw( "column select error" );
            }
         }

         delete [] param;
         delete [] value;

         UpdateModuleTable();
      }
      setEnabled(true);

   }

   return;
}



void mainform::keyPressEvent( QKeyEvent * e )
{

   if ( used_modules_table->hasFocus() )
   {
      switch ( e->key() )
      {
         case Qt::Key_Delete:
         {
            int row = used_modules_table->numRows() - 1;
            for ( vector<ExternalModule>::iterator ip = used_modules.end()-1;
                  ip != used_modules.begin()-1;
                  ip-- )
            {
               if ( used_modules_table->isRowSelected( row, true ) )
               used_modules.erase(ip);
               row--;
            }

            if ( used_modules_table->numSelections() != 0 )
            {
               reset_state();
               used_modules_table->clearSelection();
               UpdateModuleTable();
            }

            e->accept();
            return;
         }

         case Qt::Key_Insert:
         {
            ExternalModule *EM;

            QString fn = GetLoadModuleName();

            if ( fn != "" )
            {
               try
               {
                  ExternalModule *EM = new ExternalModule( (const char*)sysCodec->fromUnicode(fn) );
                  used_modules.push_back(*EM);
                  delete EM;

                  UpdateModuleTable();
                  used_modules_table->clearSelection();

//                  reset_state();
               }
               catch( const char* str )
               {
                  Xlog << "insertion error: " << str << "!" << endl;
               }
            }
            e->accept();
            return;
         }

         case Qt::Key_Space:
         {
            UsedModulesRequest( used_modules_table->currentRow(), 
                                used_modules_table->currentColumn(),
                                0,
                                QPoint(0,0) );
         }

         case Qt::Key_Escape:
         {
            used_modules_table->clearSelection();
         }

         default: {}
      }
   }

}



void mainform::getFsettings()
{
   lineEditFrequency->setText( sysLocale.toString(MD.freq*FreqUnitFactor(FreqUnit->currentText()), 'g', 10) );
   lineEditWavelength->setText( sysLocale.toString(MD.wavelength*LengthUnitFactor(wlUnit->currentText()), 'g', 10) );

   AIMsizeX->setText( sysLocale.toString(MD.size_x*LengthUnitFactor(sizeUnit->currentText()), 'g', 10) );
   AIMsizeY->setText( sysLocale.toString(MD.size_y*LengthUnitFactor(sizeUnit->currentText()), 'g', 10) );
   AIMsizeZ->setText( sysLocale.toString(MD.size_z*LengthUnitFactor(sizeUnit->currentText()), 'g', 10) );

   AIMdiscX->setText( sysLocale.toString(MD.AIMnox) );
   AIMdiscY->setText( sysLocale.toString(MD.AIMnoy) );
   AIMdiscZ->setText( sysLocale.toString(MD.AIMnoz) );

   ppwlX->setText( sysLocale.toString(MD.AIMnox*MD.wavelength/MD.size_x, 'g', 10) );
   ppwlY->setText( sysLocale.toString(MD.AIMnoy*MD.wavelength/MD.size_y, 'g', 10) );
   ppwlZ->setText( sysLocale.toString(MD.AIMnoz*MD.wavelength/MD.size_z, 'g', 10) );

   update();
}



void mainform::customEvent ( QCustomEvent *e )
{
   switch (e->type())
   {
      case APPEND_LOG_EVENT:
      {
         LogMessage->moveCursor( QTextEdit::MoveEnd, false );
         LogMessage->insert( *(const QString*)e->data() );
         delete (QString*)e->data();
         break;
      }
      case UPDATE_FSETTINGS_EVENT:
      {
         getFsettings();
         break;
      }
      case UPDATE_STATE_EVENT:
      {
         Currentstate->setSelected( S.value(), TRUE );
         break;
      }
      default:
      {
         throw( "unknown custom event" );
         break;
      }
   }
}



void mainform::next_state()
{
   setEnabled(false);
   try
   {
      S.next();
   }
   catch( const char *str )
   {
      Xlog << "cannot change state: " << str << "!" << endl;
   }
   setEnabled(true);
}

void mainform::prev_state()
{
   S.prev();
}

void mainform::reset_state()
{
   S.back_to( INITIAL );
}



void mainform::geometry_viewer()
{
   setEnabled(false);
   gv.exec_geo();
   setEnabled(true);
}



void mainform::state_effect()
{

// default
   NextState->setEnabled(true);
   PrevState->setEnabled(true);
   showGeometryAction->setEnabled(true);
   showMatrix_densityAction->setEnabled(false);
   show2D_patternAction->setEnabled(false);
   show3DpatternAction->setEnabled(false);


   switch ( S.value() )
   {
      case INITIAL:
      {
         PrevState->setEnabled(false);
         showGeometryAction->setEnabled(false);
         break;
      }
      case MAT_RDY:
      case RHS_RDY:
      case SOLVED:
      {
         showMatrix_densityAction->setEnabled(true);
         break;
      }
      case EXTRACTED:
      {
         showMatrix_densityAction->setEnabled(true);

         NextState->setEnabled(false);
         if ( SOL.ff )
         {
            show2D_patternAction->setEnabled(true);
            show3DpatternAction->setEnabled(true);
         }
         break;
      }
      default : {}
   }

}



void mainform::openExtract()
{
   setEnabled(false);
   extract_win.getMD();

   if ( extract_win.exec() == QDialog::Accepted )
   {
      S.back_to( SOLVED );
      extract_win.setMD();
   }
   setEnabled(true);
}



void mainform::open2Dpattern()
{
   setEnabled(false);
   pattern2d.exec_2dpattern();
   setEnabled(true);
}



void mainform::run_all()
{
   setEnabled(false);

   Xlog << endl;
   Xlog << "running all..." << endl;

   try
   {
      S.back_to(INITIAL);
      while( S.value() != EXTRACTED ) S.next();

      Xlog << "run all completed." << endl;
   }
   catch ( const char* str )
   {
      Xlog << "cannot change state: " << str << "!" << endl;
   }

   setEnabled(true);
}



void mainform::open3Dpattern()
{
   setEnabled(false);
   try
   {
      pattern3d.exec_3dpattern();
   }
   catch( const char *str )
   {
      Xlog << "cannot show far-field: " << str << "!" << endl;
   }
   setEnabled(true);
}



void mainform::openDensity()
{
   setEnabled(false);
   try
   {
      df.exec_density();
   }
   catch( const char *str )
   {
      Xlog << "cannot show matrix density: " << str << "!" << endl;
   }
   setEnabled(true);
}



void mainform::openHelp()
{
   QString fn = root_path
                + "/doc/html/"
                + sysLocale.name().section('_',0,0)
                + "/gantlab.html";

   if ( QFile::exists(fn) ) qac->showPage(fn);
   else                     qac->showPage( root_path + "/doc/html/gantlab.html" );
}



void mainform::openAboutQt()
{
   setEnabled(false);
   QMessageBox::aboutQt(this);
   setEnabled(true);
}



void mainform::run_rest()
{
   setEnabled(false);

   Xlog << endl;
   Xlog << "running rest..." << endl;

   try
   {
      while( S.value() != EXTRACTED ) S.next();

      Xlog << "run rest completed." << endl;
   }
   catch ( const char* str )
   {
      Xlog << "cannot change state: " << str << "!" << endl;
   }

   setEnabled(true);
}



void mainform::openAbout()
{
   About a;
   QString about_text;

   about_text = "GNU antenna laboratory\n";
   about_text += QString("version ") + VERSION + "\n";
   about_text += "\nwritten by Ingo D. Rullhusen <d01c@uni-bremen.de>\n";
   about_text += QString("\ncompiled at ") + __DATE__ + " " + __TIME__ + "\n";

   a.AboutText->setText( about_text );

   setEnabled(false);
   a.exec();
   setEnabled(true);
}



// end GUI stuff
#endif

