//
// module header file, declarations for C++ module library
// (c) 16.4.2007 Ingo D. Rullhusen
//

#ifndef MODULE_HEADER_H
#define MODULE_HEADER_H


#include <GL/gl.h>
#include <string>
#include <complex>
#include "maindata.h"
#include "field.h"


//
// MODULE HEADER

#define module_header_version_ -123
extern "C" int MainData_version;      // must be set to MainData_version_
extern "C" int FIELD_version;         // must be set to FIELD_version_
extern "C" int module_header_version; // must be set to module_header_version_


extern "C" int version;   // must be set to module version
extern "C" char *name;    // must be set to module name

// internal data handler
extern "C" void* new_data();
extern "C" void* dpl_data( void* data );
extern "C" void  del_data( void* data );


//
// GEOMETRY & MATRIX

// get and set module parameters for geometry and matrix
extern "C" void matrix_get_parameter( const MainData *MD, void *data,
                                      string **param, string **value,
                                      int *no );
// new param & value
extern "C" void matrix_set_parameter( const MainData *MD, void *data,
                                      const string *param, const string *value,
                                      int no );

// geometrical set-up stuff
extern "C" int  matrix_geo_setup( const MainData *MD, void *data,
                                  string **omsg, int *nomsg );
// new omsg
// return value: no. of bfkt
extern "C" void matrix_renumber( const MainData *MD, void *data, int *perm, int *rperm );
extern "C" void matrix_bounding_box( const MainData *MD, void *data,
                                     double *cpos_x, double *cpos_y, double *cpos_z,
                                     double *size_x, double *size_y, double *size_z );
extern "C" void matrix_add_to_OGLlist( const MainData *MD, void *data, GLuint gl_list );

// AIM stuff for basis (nbfkt<nobfkt) and excitation (nbfkt>=nobfkt)
// allocate subdomain list
extern "C" void matrix_bfkt_bounding_box( const MainData *MD, void *data, int nbfkt,
                                          int *subno,
                                          double **cpos_x, double **cpos_y, double **cpos_z,
                                          double **size_x, double **size_y, double **size_z );
extern "C" void matrix_bfkt_moment( const MainData *MD, void *data, int nbfkt, int nsub,
                                    double cpos_x, double cpos_y, double cpos_z,
                                    double len_x, double len_y, double len_z,
                                    int nx, int ny, int nz,
                                    double *mx, double *my, double *mz, double *md );

// matrix entries
// EFIE: E{J} + E{J^inc} + E^inc = 0
extern "C" void matrix_ele_setup( const MainData *MD, void *data,
                                  string **omsg, int *nomsg );
// new omsg
extern "C" complex<double> matrix_ele( const MainData *MD, void *data, int i, int j );
// return value: + Integral bfkt_i E{ bfkt_j }
extern "C" void matrix_ele_cleanup( const MainData *MD, void *data );


//
// RHS

// get and set module parameters for right hand side
extern "C" void RHS_get_parameter( const MainData *MD, void *data,
                                   string **param, string **value,
                                   int *no );
// new param & value
extern "C" void RHS_set_parameter( const MainData *MD, void *data,
                                   const string *param, const string *value,
                                   int no );

extern "C" void RHS_add_to_OGLlist( const MainData *MD, void *data, GLuint gl_list );

// RHS entries
// EFIE: E{J} + E{J^inc} + E^inc = 0
extern "C" void RHS_E_setup( const MainData *MD, void *data,
                             string **omsg, int *nomsg );
// new omsg
extern "C" void RHS_Einc( const MainData *MD, void *data, double x, double y, double z,
                          complex<double> E[3] );
// non-current E only
// E[3] = + E^inc
extern "C" complex<double> RHS_E_internal( const MainData *MD, void *data, int i );
// see AIM
// return value: + Integral bfkt_i E{J^inc}
extern "C" void RHS_bfkt_inc( const MainData *MD, void *data,
                              complex<double> **ampl, int *no );
// new ampl, no appended to nobfkt
extern "C" void RHS_E_cleanup( const MainData *MD, void *data );


//
// EXTRACT

// get and set module parameters for result extraction
extern "C" void extract_get_parameter( const MainData *MD, void *data,
                                       string **param, string **value,
                                       int *no );
// new param & value
extern "C" void extract_set_parameter( const MainData *MD, void *data,
                                       const string *param, const string *value,
                                       int no );

extern "C" void extract_add_to_OGLlist( const MainData *MD, void *data, GLuint gl_list );
extern "C" void extract_run( const MainData *MD, void *data, bool intern_inc,
                             complex<double> *JV, int noJ, const FIELD *f,
                             string **omsg, int *nomsg );
// new omsg


#endif
