//
// module interface class
// (c) 16.4.2007 Ingo D. Rullhusen
//

#include <dlfcn.h>

#include "global_vars.h"

#include "module_interface.h"
#include "auxiliary.h"
#include "module_header.h"



void* new_data_dummy()
{ return(NULL); }

void* dpl_data_dummy( void *data )
{ return(NULL); }

void del_data_dummy( void *data )
{}


void get_parameter_dummy( const MainData *MD, void *d, string **p, string **v, int *n )
{
   *n = 0;
   *p = NULL;
   *v = NULL;
}

void set_parameter_dummy( const MainData *MD, void *d, const string *p, const string *v, int n)
{}

int geo_setup_dummy( const MainData *MD, void *d, string **omsg, int *nomsg )
{ *nomsg=0; *omsg=NULL; return(0); }

void setup_dummy( const MainData *MD, void *d, string **omsg, int *nomsg )
{ *nomsg=0; *omsg=NULL; }

void renumber_dummy( const MainData *MD, void *d, int *perm, int *rperm )
{}

void bounding_box_dummy( const MainData *MD, void *d,
                         double *cpos_x, double *cpos_y, double *cpos_z,
                         double *size_x, double *size_y, double *size_z )
{
   *cpos_x = MD->cpos_x;
   *cpos_y = MD->cpos_y;
   *cpos_z = MD->cpos_z;
   
   *size_x = *size_y = *size_z = 0.0;
}

void add_to_OGLlist_dummy( const MainData *MD, void *d, GLuint gl )
{
   
}

void cleanup_dummy( const MainData *MD, void *d )
{}

void bfkt_inc_dummy( const MainData *MD, void *data, complex<double> **ampl, int *no )
{
   *no = 0;
   *ampl = NULL;
}

void extract_run_dummy( const MainData *MD, void *d, bool internal_inc,
                        complex<double> *JV, int no, const FIELD *f,
                        string **omsg, int *nomsg )
{
   *nomsg = 0;
   *omsg = NULL;
}


bool ExternalModule::init( string fn )
{
   new_data_ = new_data_dummy;
   dpl_data_ = dpl_data_dummy;
   del_data_ = del_data_dummy;

   handle = dlopen(fn.c_str(), RTLD_NOW);

   if ( handle == NULL )
   {
      char *dlerr = dlerror();

      Xlog << "load problem: " << sysCodec->toUnicode(dlerr) << endl;

      return(false);
   }

   int *mdv, *fv, *hdrv;

   mdv  = (int*)dlsym( handle, "MainData_version" );
   fv   = (int*)dlsym( handle, "FIELD_version" );
   hdrv = (int*)dlsym( handle, "module_header_version" );

   bool verr = true;
   if ( mdv )
      if ( *mdv == MainData_version_ )
         if ( fv )
            if ( *fv == FIELD_version_ )
               if ( hdrv )
                  if ( *hdrv == module_header_version_ )
                  { verr = false; }

   if ( verr )
   {
      Xlog << "load problem: version mismatch"<< endl;
      dlclose(handle);
      handle = NULL;
      return(false);
   }



   modfile_ = fn;
   
   
   version_ = (int*)dlsym( handle, "version" );
   name_ = *((char**)dlsym( handle, "name" ));
   
   new_data_ = (void* (*)())
               dlsym( handle, "new_data" );
   dpl_data_ = (void* (*)( void* ))
               dlsym( handle, "dpl_data" );
   del_data_ = (void (*)( void* ))
               dlsym( handle, "del_data" );
   
   if ( !(version_&&name_&&new_data_&&dpl_data_&&del_data_) )
   {
      dlclose(handle);
      return(false);
   }
   
   
   matrix_get_parameter_ = (void (*)( const MainData *, void *,
                                      string **, string **, int * ))
   dlsym( handle, "matrix_get_parameter");
   matrix_set_parameter_ = (void (*)( const MainData *, void *,
                                      const string *, const string *, int ))
   dlsym( handle, "matrix_set_parameter");
   
   matrix_geo_setup_ = (int (*)( const MainData *, void *, string **, int * ))
                       dlsym( handle, "matrix_geo_setup");
   matrix_renumber_ = (void (*)( const MainData *, void *, int *, int * ))
                      dlsym( handle, "matrix_renumber");
   matrix_bounding_box_ = (void (*)( const MainData *, void *,
                                     double *, double *, double *,
                                     double *, double *, double * ))
   dlsym( handle, "matrix_bounding_box");

   matrix_add_to_OGLlist_ = (void (*)( const MainData *, void *, GLuint ))
                            dlsym( handle, "matrix_add_to_OGLlist");
   
   matrix_bfkt_bounding_box_ = (void (*)( const MainData *, void *, int,
                                          int *,
                                          double **, double **, double **,
                                          double **, double **, double **  ))
   dlsym( handle, "matrix_bfkt_bounding_box");

   matrix_bfkt_moment_ = (void (*)( const MainData *, void *, int, int,
                                    double, double, double,
                                    double, double, double,
                                    int, int, int,
                                    double *, double *, double *, double * ))
   dlsym( handle, "matrix_bfkt_moment");
   
   
   matrix_ele_setup_ = (void (*)( const MainData *, void *, string **, int * ))
                       dlsym( handle, "matrix_ele_setup");
   matrix_ele_ = (complex<double> (*)( const MainData *, void *, int, int ))
                 dlsym( handle, "matrix_ele");
   matrix_ele_cleanup_ = (void (*)( const MainData *, void * ))
                         dlsym( handle, "matrix_ele_cleanup");

   
   has_matrix_ = matrix_geo_setup_ || matrix_add_to_OGLlist_ ||
                 ( matrix_renumber_
                   && matrix_bfkt_bounding_box_ && matrix_bfkt_moment_
                   && matrix_ele_ );
   
   if ( ! matrix_get_parameter_ ) matrix_get_parameter_ = get_parameter_dummy;
   if ( ! matrix_set_parameter_ ) matrix_set_parameter_ = set_parameter_dummy;
   
   if ( ! matrix_geo_setup_ ) matrix_geo_setup_ = geo_setup_dummy;
   if ( ! matrix_renumber_ ) matrix_renumber_ = renumber_dummy;
   if ( ! matrix_bounding_box_ ) matrix_bounding_box_ = bounding_box_dummy;
   if ( ! matrix_add_to_OGLlist_ ) matrix_add_to_OGLlist_ = add_to_OGLlist_dummy;
   if ( ! matrix_ele_setup_ ) matrix_ele_setup_ = setup_dummy;
   if ( ! matrix_ele_cleanup_ ) matrix_ele_cleanup_ = cleanup_dummy;




   RHS_get_parameter_ = (void (*)( const MainData *, void *,
                                   string **, string **, int * ))
   dlsym( handle, "RHS_get_parameter");
   RHS_set_parameter_ = (void (*)( const MainData *, void *,
                                   const string *, const string *, int ))
   dlsym( handle, "RHS_set_parameter");
   
   RHS_add_to_OGLlist_ = (void (*)( const MainData *, void *, GLuint ))
                         dlsym( handle, "RHS_add_to_OGLlist");
   
   RHS_E_setup_ = (void (*)( const MainData *, void *, string **, int * ))
                  dlsym( handle, "RHS_E_setup");
   
   RHS_Einc_ = (void (*)( const MainData *, void *, double, double, double, complex<double> [3] ))
               dlsym( handle, "RHS_Einc" );

   RHS_bfkt_inc_ = (void (*)( const MainData *, void *, complex<double> **, int * ))
                   dlsym( handle, "RHS_bfkt_inc" );

   RHS_E_internal_ = (complex<double> (*)( const MainData *, void *, int ))
                     dlsym( handle, "RHS_E_internal");

   RHS_E_cleanup_ = (void (*)( const MainData *, void * ))
                    dlsym( handle, "RHS_E_cleanup");
   
   has_RHS_ = RHS_add_to_OGLlist_ || RHS_Einc_;
   
   if ( ! RHS_get_parameter_ ) RHS_get_parameter_ = get_parameter_dummy;
   if ( ! RHS_set_parameter_ ) RHS_set_parameter_ = set_parameter_dummy;
   
   if ( ! RHS_add_to_OGLlist_ ) RHS_add_to_OGLlist_ = add_to_OGLlist_dummy;
   
   if ( ! RHS_E_setup_ ) RHS_E_setup_ = setup_dummy;
   if ( ! RHS_E_cleanup_ ) RHS_E_cleanup_ = cleanup_dummy;

   if ( ! RHS_bfkt_inc_ ) RHS_bfkt_inc_ = bfkt_inc_dummy;




   extract_get_parameter_ = (void (*)( const MainData *, void *,
                                       string **, string **, int * ))
   dlsym( handle, "extract_get_parameter");
   extract_set_parameter_ = (void (*)( const MainData *, void *,
                                       const string *, const string *, int ))
   dlsym( handle, "extract_set_parameter");
   
   extract_add_to_OGLlist_ = (void (*)( const MainData *, void *, GLuint ))
                             dlsym( handle, "extract_add_to_OGLlist");
   
   
   extract_run_ = (void (*)( const MainData *, void *, bool,
                             complex<double> *, int, const FIELD *,
                             string **, int * ))
   dlsym( handle, "extract_run");
   
   
   has_extract_ = extract_get_parameter_ || extract_set_parameter_
                  || extract_add_to_OGLlist_ || extract_run_;
   
   if ( ! extract_get_parameter_ ) extract_get_parameter_ = get_parameter_dummy;
   if ( ! extract_set_parameter_ ) extract_set_parameter_ = set_parameter_dummy;
   
   if ( ! extract_add_to_OGLlist_ ) extract_add_to_OGLlist_ = add_to_OGLlist_dummy;
   
   if ( ! extract_run_ ) extract_run_ = extract_run_dummy;
   
   
   no_of_bfkt_ = matrix_offset_ = 0;


   return(true);
}



ExternalModule::ExternalModule( string fn )
{
   if ( ! init( fn ) ) throw( "module load error" );
   use_matrix_ = use_RHS_ = use_extract_ = false;
   no_of_bfkt_ = matrix_offset_ = 0;
   data = new_data_();
}


ExternalModule::ExternalModule( const ExternalModule &EM )
{
   if ( ! init( EM.modfile() ) ) throw( "module load error" );
   
   use_matrix_ = EM.use_matrix_;
   use_RHS_ = EM.use_RHS_;
   use_extract_ = EM.use_extract_;
   
   no_of_bfkt_ = EM.no_of_bfkt_;
   matrix_offset_ = EM.matrix_offset_;
   
   data = dpl_data_( EM.data );
}


ExternalModule::~ExternalModule()
{
   del_data_( data );
   data = NULL;
   
   dlclose( handle );
   handle = NULL;
}

ExternalModule& ExternalModule::operator =( const ExternalModule rhs )
{
   del_data_( data );
   dlclose( handle );
   if ( ! init( rhs.modfile() ) ) throw( "module load error" );
   
   use_matrix_ = rhs.use_matrix_;
   use_RHS_ = rhs.use_RHS_;
   use_extract_ = rhs.use_extract_;
   
   no_of_bfkt_ = rhs.no_of_bfkt_;
   matrix_offset_ = rhs.matrix_offset_;
   
   data = dpl_data_( rhs.data );
   
   return(*this);
}


void ExternalModule::write_parameter( const MainData *MD, fstream &file )
{
   file << "[ExternalModule]" << endl;
   
   file << "\"" << modfile().substr( modfile().rfind('/')+1 ) << "\""
         << "   #module file" << endl;
   
   int no;
   string *param_t, *value_t;
   
   matrix_get_parameter_( MD, data, &param_t, &value_t, &no );
   
   file << no << " " << use_matrix_ << "   #no_matrix_para used" << endl;
   for ( int i=0; i<no; i++ )
   {
      file << "\"" << param_t[i] << "\" \"" << value_t[i] << "\"" << endl;
   }
   delete [] param_t;
   delete [] value_t;
   
   
   RHS_get_parameter_( MD, data, &param_t, &value_t, &no );
   
   file << no << " " << use_RHS_ << "   #no_RHS_para used" << endl;
   for ( int i=0; i<no; i++ )
   {
      file << "\"" << param_t[i] << "\" \"" << value_t[i] << "\"" << endl;
   }
   delete [] param_t;
   delete [] value_t;
   
   
   extract_get_parameter_( MD, data, &param_t, &value_t, &no );
   
   file << no << " " << use_extract_ << "   #no_extract_para used" << endl;
   for ( int i=0; i<no; i++ )
   {
      file << "\"" << param_t[i] << "\" \"" << value_t[i] << "\"" << endl;
   }
   delete [] param_t;
   delete [] value_t;
}

ExternalModule::ExternalModule( const MainData *MD, fstream &file )
{
   int no;
   string sbuff = "";

   no_of_bfkt_ = 0;
   matrix_offset_ = 0;

   while ( file.good() && sbuff != "ExternalModule" )
   { sbuff = get_section_label( file, true ); }
   
   if ( sbuff != "ExternalModule" ) throw( "cannot find ExternalModule entry" );
   
   sbuff = get_quoted_string( file );
   goto_next_line( file );
   
   if ( ! init( MD->ModulePath+"/"+sbuff ) ) throw( "module load error (path?)" );
   
   data = new_data_();
   
   // matrix
   file >> no >> use_matrix_;
   goto_next_line( file );
   
   string *param_t = new string[no],
   *value_t = new string[no];
   
   for ( int i=0; i<no; i++ )
   {
      param_t[i] = get_quoted_string( file );
      value_t[i] = get_quoted_string( file );
      goto_next_line( file );
   }
   matrix_set_parameter_( MD, data, param_t, value_t, no );
   
   delete [] value_t;
   delete [] param_t;
   
   // RHS
   file >> no >> use_RHS_;
   goto_next_line( file );
   
   param_t = new string[no];
   value_t = new string[no];
   
   for ( int i=0; i<no; i++ )
   {
      param_t[i] = get_quoted_string( file );
      value_t[i] = get_quoted_string( file );
      goto_next_line( file );
   }
   RHS_set_parameter_( MD, data, param_t, value_t, no );
   
   delete [] value_t;
   delete [] param_t;
   
   // extract
   file >> no >> use_extract_;
   goto_next_line( file );
   
   param_t = new string[no];
   value_t = new string[no];
   
   for ( int i=0; i<no; i++ )
   {
      param_t[i] = get_quoted_string( file );
      value_t[i] = get_quoted_string( file );
      goto_next_line( file );
   }
   extract_set_parameter_( MD, data, param_t, value_t, no );
   
   delete [] value_t;
   delete [] param_t;
   
}


void ExternalModule::cleanup( const MainData *MD )
{
   int no1, no2, no3;
   string *param1_t, *value1_t,
          *param2_t, *value2_t,
          *param3_t, *value3_t;
   
   matrix_get_parameter_( MD, data, &param1_t, &value1_t, &no1 );
   RHS_get_parameter_( MD, data, &param2_t, &value2_t, &no2 );
   extract_get_parameter_( MD, data, &param3_t, &value3_t, &no3 );

   del_data_( data );
   data = new_data_();

   matrix_set_parameter_( MD, data, param1_t, value1_t, no1 );
   RHS_set_parameter_( MD, data, param2_t, value2_t, no2 );
   extract_set_parameter_( MD, data, param3_t, value3_t, no3 );

   delete [] param1_t;
   delete [] value1_t;
   delete [] param2_t;
   delete [] value2_t;
   delete [] param3_t;
   delete [] value3_t;

   no_of_bfkt_ = 0;
   matrix_offset_ = 0;
}
