//
// module interface class
// (c) 26.3.2007 Ingo D. Rullhusen
//

#ifndef MODULE_INTERFACE_H
#define MODULE_INTERFACE_H

#include <fstream>
#include <complex>

#include <GL/gl.h>

#include <string.h>

#include "maindatax.h"
#include "field.h"


class ExternalModule
{
   public:
      string modfile() const{ return(modfile_); };
      int version() const{ return(*version_); };
      string name() const{ string s = name_; return(s); };


      bool has_matrix() const{ return(has_matrix_); };
      bool has_RHS() const{ return(has_RHS_); };
      bool has_extract() const{ return(has_extract_); };

      bool& use_matrix(){ return(use_matrix_); };
      bool& use_RHS(){ return(use_RHS_); };
      bool& use_extract(){ return(use_extract_); };

      bool has_E_internal(){ return(RHS_E_internal_); };
      bool has_Einc(){ return(RHS_Einc_); };


      void matrix_get_parameter( const MainData *MD,   // new param & value
                                 string **param, string **value,
                                 int *no ) const
      { matrix_get_parameter_( MD, data, param, value, no ); };

      void matrix_set_parameter( const MainData *MD,
                                 const string *param, const string *value,
                                 int no )
      { matrix_set_parameter_( MD, data, param, value, no ); };

      void matrix_geo_setup( const MainData *MD, string **omsg, int *nomsg )
      { no_of_bfkt_ = matrix_geo_setup_( MD, data, omsg, nomsg ); };

      void matrix_renumber( const MainData *MD, int *perm, int *rperm )
      { matrix_renumber_( MD, data, perm, rperm ); };

      void matrix_bounding_box( const MainData *MD,
                                double *cpos_x, double *cpos_y, double *cpos_z,
                                double *size_x, double *size_y, double *size_z ) const
      { matrix_bounding_box_( MD, data, cpos_x, cpos_y, cpos_z, size_x, size_y, size_z ); };

      void matrix_add_to_OGLlist( const MainData *MD, GLuint gl_list ) const
      { matrix_add_to_OGLlist_( MD, data, gl_list ); };


      int no_of_bfkt() const { return(no_of_bfkt_); };
      int& matrix_offset(){ return(matrix_offset_); };


      void matrix_bfkt_bounding_box( const MainData *MD, int nbfkt, int *subno,
                                     double **cpos_x, double **cpos_y, double **cpos_z,
                                     double **size_x, double **size_y, double **size_z ) const
      { matrix_bfkt_bounding_box_( MD, data, nbfkt, subno,
                                   cpos_x, cpos_y, cpos_z,
                                   size_x, size_y, size_z ); };
      void matrix_bfkt_moment( const MainData *MD, int nbfkt, int nsub,
                               double cpos_x, double cpos_y, double cpos_z,
                               double len_x, double len_y, double len_z,
                               int nx, int ny, int nz,
                               double *mx, double *my, double *mz, double *md ) const
      { matrix_bfkt_moment_( MD, data, nbfkt, nsub,
                             cpos_x, cpos_y, cpos_z, len_x, len_y, len_z,
                             nx, ny, nz, mx, my, mz, md ); };


      void matrix_ele_setup( const MainData *MD, string **omsg, int *nomsg )
      { matrix_ele_setup_( MD, data, omsg, nomsg ); };
      complex<double> matrix_ele( const MainData *MD, int i, int j ) const
      { return( matrix_ele_( MD, data, i, j ) ); };
      void matrix_ele_cleanup( const MainData *MD ){ matrix_ele_cleanup_( MD, data ); };




      void RHS_get_parameter( const MainData *MD,   // new param & value
                              string **param, string **value,
                              int *no )
      { RHS_get_parameter_( MD, data, param, value, no ); };

      void RHS_set_parameter( const MainData *MD,
                              const string *param, const string *value,
                              int no )
      { RHS_set_parameter_( MD, data, param, value, no ); };

      void RHS_add_to_OGLlist( const MainData *MD, GLuint gl_list )
      { RHS_add_to_OGLlist_( MD, data, gl_list ); };

      void RHS_E_setup( const MainData *MD, string **omsg, int *nomsg )
      { RHS_E_setup_( MD, data, omsg, nomsg ); };
      void RHS_Einc( const MainData *MD, double x, double y, double z, complex<double> E[3] )
      { RHS_Einc_( MD, data, x, y, z, E ); };
      complex<double> RHS_E_internal( const MainData *MD, int i )
      { return( RHS_E_internal_( MD, data, i ) ); };
      void RHS_bfkt_inc( const MainData *MD, complex<double> **ampl, int *no )
      { RHS_bfkt_inc_( MD, data, ampl, no ); };
      void RHS_E_cleanup( const MainData *MD ){ RHS_E_cleanup_( MD, data ); };




      void extract_get_parameter( const MainData *MD,   // new param & value
                                  string **param, string **value,
                                  int *no )
      { extract_get_parameter_( MD, data, param, value, no ); };

      void extract_set_parameter( const MainData *MD,
                                  const string *param, const string *value,
                                  int no )
      { extract_set_parameter_( MD, data, param, value, no ); };

      void extract_add_to_OGLlist( const MainData *MD, GLuint gl_list )
      { extract_add_to_OGLlist_( MD, data, gl_list ); };

      
      void extract_run( const MainData *MD, complex<double> *JV, int noJ, const FIELD *f,
                        string **omsg, int *nomsg )
      { extract_run_( MD, data, use_RHS_, JV, noJ, f, omsg, nomsg ); };



      void cleanup( const MainData *MD );
      void write_parameter( const MainData *MD, fstream &file );

      ExternalModule( string modfile );
      ExternalModule( const ExternalModule &EM );
      ExternalModule( const MainData *MD, fstream &file );
      ~ExternalModule();

      ExternalModule& operator =( const ExternalModule rhs );

   private:
      bool init( const string );

      bool use_matrix_, use_RHS_, use_extract_;

      int no_of_bfkt_, matrix_offset_;

      string modfile_;
      void *handle;

// module contents
      int *version_;
      char *name_;

      void* (*new_data_)();
      void* (*dpl_data_)( void *data );
      void  (*del_data_)( void *data );

      bool has_matrix_, has_RHS_, has_extract_;


      void (*matrix_get_parameter_)( const MainData *MD, void *data,
                                     string **param, string **value,
                                     int *no );
      void (*matrix_set_parameter_)( const MainData *MD, void *data,
                                     const string *param, const string *value,
                                     int no );


      int (*matrix_geo_setup_)( const MainData *MD, void *data,
                                string **omsg, int *nomsg );
      void (*matrix_renumber_)( const MainData *MD, void *data, int *perm, int *rperm );
      void (*matrix_bounding_box_)( const MainData *MD, void *data,
                                    double *cpos_x, double *cpos_y, double *cpos_z,
                                    double *size_x, double *size_y, double *size_z );
      void (*matrix_add_to_OGLlist_)( const MainData *MD, void *data, GLuint gl_list );


      void (*matrix_bfkt_bounding_box_)( const MainData *MD, void *data, int nbfkt,
                                         int *subno,
                                         double **cpos_x, double **cpos_y, double **cpos_z,
                                         double **size_x, double **size_y, double **size_z );
      void (*matrix_bfkt_moment_)( const MainData *MD, void *data, int nbfkt, int nsub,
                                   double cpos_x, double cpos_y, double cpos_z,
                                   double len_x, double len_y, double len_z,
                                   int nx, int ny, int nz,
                                   double *mx, double *my, double *mz, double *md );


      void (*matrix_ele_setup_)( const MainData *MD, void *data, string **omsg, int *nomsg );
      complex<double> (*matrix_ele_)( const MainData *MD, void *data, int i, int j );
      void (*matrix_ele_cleanup_)( const MainData *MD, void *data );



      void (*RHS_get_parameter_)( const MainData *MD, void *data,
                                  string **param, string **value,
                                  int *no );
      void (*RHS_set_parameter_)( const MainData *MD, void *data,
                                  const string *param, const string *value,
                                  int no );
      void (*RHS_add_to_OGLlist_)( const MainData *MD, void *data, GLuint gl_list );

      void (*RHS_E_setup_)( const MainData *MD, void *data, string **omsg, int *nomsg );
      void (*RHS_Einc_)( const MainData *MD, void *data, double x, double y, double z,
                         complex<double> E[3] );
      complex<double> (*RHS_E_internal_)( const MainData *MD, void *data, int i );
      void (*RHS_bfkt_inc_)( const MainData *MD, void *data,
                             complex<double> **ampl, int *no );
      void (*RHS_E_cleanup_)( const MainData *MD, void *data );



      void (*extract_get_parameter_)( const MainData *MD, void *data,
                                      string **param, string **value,
                                      int *no );
      void (*extract_set_parameter_)( const MainData *MD, void *data,
                                      const string *param, const string *value,
                                      int no );
      void (*extract_add_to_OGLlist_)( const MainData *MD, void *data, GLuint gl_list );

      void (*extract_run_)( const MainData *MD, void *data, bool internal_inc,
                            complex<double> *JV, int noJ, const FIELD *f,
                            string **omsg, int *nomsg );



// data of instance
      void *data;
};

#endif
