//
// OpenGL Widget (3D pattern)
// (c) 10.11.2007 Ingo D. Rullhusen
//


#include <float.h>
#include <math.h>


#include <qprogressdialog.h>
#ifndef TERMINAL
#include <qfiledialog.h>
#include <qimage.h>
#endif


#include "global_vars.h"

#ifdef TERMINAL
#include "terminal/ogl3dpattern.h"
#else
#include "ogl3dpattern.h"
#endif
#include "auxiliary.h"
#include "rotmat.h"


#define MAX_OGL_OBJ 5

#ifdef TERMINAL
OGL3dpattern::OGL3dpattern()
#else
OGL3dpattern::OGL3dpattern( QWidget* parent, const char* name )
: OGLdrawing::OGLdrawing( parent, name )
#endif
{
#ifndef TERMINAL
   makeCurrent();
   partlist = glGenLists( MAX_OGL_OBJ );
#endif

   coordinates = tr("(none)");
   notheta = nophi = 0;
   RefValue = CrossValue = NULL;
   MaxValue = 0.0;

   lvlM = false;
   mpy = 0;
   rmax = rmin = 0.0;

   clearall();
}

OGL3dpattern::~OGL3dpattern()
{
#ifndef TERMINAL
   makeCurrent();
   glDeleteLists( partlist, MAX_OGL_OBJ );
#endif

   delete [] RefValue;
   delete [] CrossValue;
}


void OGL3dpattern::clearall()
{
#ifndef TERMINAL
   makeCurrent();
   glDeleteLists( object, 1 );
   object = glGenLists( 1 );
#endif
   R = 0.75*sqrt( MD.size_x*MD.size_x
                  + MD.size_y*MD.size_y
                  + MD.size_z*MD.size_z );
}


void OGL3dpattern::init_ff( FARFIELD *FF, QString cosys )
{
   notheta = MD.no_theta;
   nophi = MD.no_phi + 1;  // last=first

   if ( FF == NULL ) throw( "far-field data not present" );

   delete [] RefValue;
   delete [] CrossValue;

   RefValue = new complex<double>[ notheta*nophi ];
   CrossValue = new complex<double>[ notheta*nophi ];

   int idx;
   double ref[3], cross[3], r[3];
   double lth, lphi, cosphi, sinphi, costh, sinth;
   complex<double> E[3];
   QProgressDialog *PW;

   PW = new QProgressDialog( QProgressDialog::tr("initialize far-field..."),
                             QProgressDialog::tr("Cancel"),
                             notheta,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(0);

   try
   {

   for ( int ith=0; ith<notheta; ith++ )
   {
      lth = ith*M_PI / (notheta-1);
      sinth = sin(lth);
      costh = cos(lth);


      PW->setProgress( ith+1 );


      for ( int iphi=0; iphi<nophi; iphi++ )
      {

         if ( PW->wasCanceled() )
         {
            throw( "far-field initialization canceled" );
         }
         ptrAPP->processEvents();


         lphi = iphi*2.0*M_PI / (nophi-1);
         sinphi = sin(lphi);
         cosphi = cos(lphi);
         

         r[0] = lx[0]*sinth*cosphi + ly[0]*sinth*sinphi + lz[0]*costh;
         r[1] = lx[1]*sinth*cosphi + ly[1]*sinth*sinphi + lz[1]*costh;
         r[2] = lx[2]*sinth*cosphi + ly[2]*sinth*sinphi + lz[2]*costh;

         FF->E_ff( E, r[0], r[1], r[2], MD.itpl_deg );


         if ( cosys == tr("spherical") )
         {
            ref[0] = lx[0]*costh*cosphi + ly[0]*costh*sinphi - lz[0]*sinth;
            ref[1] = lx[1]*costh*cosphi + ly[1]*costh*sinphi - lz[1]*sinth;
            ref[2] = lx[2]*costh*cosphi + ly[2]*costh*sinphi - lz[2]*sinth;

            cross[0] = -lx[0]*sinphi + ly[0]*cosphi;
            cross[1] = -lx[1]*sinphi + ly[1]*cosphi;
            cross[2] = -lx[2]*sinphi + ly[2]*cosphi;
         }
         else if ( cosys == tr("Ludwig 3") )
         {
            double phi_l = atan2( costh, -sinth*cosphi );
            double theta_l = acos( sinth*sinphi );

            double sinphi_l = sin( phi_l );
            double cosphi_l = cos( phi_l );
            double sinth_l = sin( theta_l );
            double costh_l = cos( theta_l );

            double thv[3], phv[3];

            thv[0] = -lx[0]*costh_l*cosphi_l + lz[0]*costh_l*sinphi_l - ly[0]*sinth_l;
            thv[1] = -lx[1]*costh_l*cosphi_l + lz[1]*costh_l*sinphi_l - ly[1]*sinth_l;
            thv[2] = -lx[2]*costh_l*cosphi_l + lz[2]*costh_l*sinphi_l - ly[2]*sinth_l;

            phv[0] = lx[0]*sinphi_l + lz[0]*cosphi_l;
            phv[1] = lx[1]*sinphi_l + lz[1]*cosphi_l;
            phv[2] = lx[2]*sinphi_l + lz[2]*cosphi_l;


            ref[0] = sinphi_l * thv[0] + cosphi_l * phv[0];
            ref[1] = sinphi_l * thv[1] + cosphi_l * phv[1];
            ref[2] = sinphi_l * thv[2] + cosphi_l * phv[2];

            cross[0] = cosphi_l * thv[0] - sinphi_l * phv[0];
            cross[1] = cosphi_l * thv[1] - sinphi_l * phv[1];
            cross[2] = cosphi_l * thv[2] - sinphi_l * phv[2];
         }
         else if ( cosys == tr("cartesian") )
         {
            ref[0] = lz[0];
            ref[1] = lz[1];
            ref[2] = lz[2];

            cross[0] = lx[0];
            cross[1] = lx[1];
            cross[2] = lx[2];
         }
         else
            throw( "unknown coordinate system" );


         idx = iphi + nophi*ith;
         RefValue[ idx ]   = E[0]*ref[0]   + E[1]*ref[1]   + E[2]*ref[2];
         CrossValue[ idx ] = E[0]*cross[0] + E[1]*cross[1] + E[2]*cross[2];
      }
   }

   }
   catch( const char *str )
   {
      notheta = nophi = 0;
      delete [] RefValue;
      RefValue = NULL;
      delete [] CrossValue;
      CrossValue = NULL;

      PW->cancel();
      delete PW;

      throw(str);
   }

   delete PW;


   MaxValue = abs( RefValue[0] );
   for ( int i=1; i<notheta*nophi; i++ )
   {
      if ( MaxValue < abs( RefValue[i] ) ) MaxValue = abs( RefValue[i] );
   }

   coordinates = cosys;
}


void OGL3dpattern::write_ffsph( fstream &file ) const
{
   if ( coordinates != tr("spherical") )
   {
      Xlog << "coordinates must be spherical!" << endl;
      file.setstate( ios::failbit );
   }

   int idx;

   file << "#FFsph" << endl;
   file << "#created by gANTlab" << endl;
   file << MD.freq << endl;
   file << notheta << " " << nophi << endl;
   file << endl;

   for ( int j=0; j<notheta; j++ )
   {
      for ( int i=0;  i<nophi; i++ )
      {
         idx = i + j*nophi;

         file << RefValue[idx].real() << " " << RefValue[idx].imag() << "  ";
      }
      file << endl;
   }

   file << endl;

   for ( int j=0; j<notheta; j++ )
   {
      for ( int i=0;  i<nophi; i++ )
      {
         idx = i + j*nophi;

         file << CrossValue[idx].real() << " " << CrossValue[idx].imag() << "  ";
      }
      file << endl;
   }

}


void OGL3dpattern::setup_refplane( double theta, double phi, double psi )
{
#ifndef TERMINAL
   const GLfloat clr_plane_ref[4]   = { 0.70, 0.00, 0.00, 0.30 };
   const GLfloat clr_plane_cross[4] = { 0.20, 0.20, 0.70, 0.40 };
   const GLfloat clr_plane_x[4]   = { 0.00, 1.00, 0.00, 1.00 };
   const GLfloat clr_plane_y[4]   = { 1.00, 1.00, 0.00, 1.00 };
   const GLfloat clr_plane_z[4]   = { 1.00, 0.00, 0.00, 1.00 };
   GLuint gl_list;
#endif
   const int noscanplp = 30;
   double rmat[3][3], sw;


   rotmat( rmat, 1.0, 0.0, 0.0, psi, theta, phi );   // dipole/ref global x -> local z
   lx[0] = rmat[0][1];
   lx[1] = rmat[1][1];
   lx[2] = rmat[2][1];

   ly[0] = rmat[0][2];
   ly[1] = rmat[1][2];
   ly[2] = rmat[2][2];

   lz[0] = rmat[0][0];
   lz[1] = rmat[1][0];
   lz[2] = rmat[2][0];


#ifndef TERMINAL
   makeCurrent();

   gl_list = partlist + 3;
   glNewList( gl_list, GL_COMPILE );

   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );


   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_ref );

   glBegin(GL_POLYGON);

   glNormal3d( lx[0], lx[1], lx[2] );
   glVertex3d( R*ly[0], R*ly[1], R*ly[2] );

   for ( int i=1; i<noscanplp; i++ )
   {
      sw = i*2.0*M_PI/noscanplp;

      rotmat( rmat, R, 0.0, sw, psi, theta, phi );

      glVertex3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   }

   glEnd();


   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_cross );

   glBegin(GL_POLYGON);

   glNormal3d( lz[0], lz[1], lz[2] );
   glVertex3d( R*ly[0], R*ly[1], R*ly[2] );

   for ( int i=1; i<noscanplp; i++ )
   {
      sw = i*2.0*M_PI/noscanplp;

      rotmat( rmat, R, 0.0, sw, psi+0.5*M_PI, theta, phi );

      glVertex3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   }

   glEnd();


   glDisable( GL_BLEND );


   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_x );

   glBegin(GL_LINES);
   glVertex3d( 1.01*R*lx[0], 1.01*R*lx[1], 1.01*R*lx[2] );
   glVertex3d( 1.10*R*lx[0], 1.10*R*lx[1], 1.10*R*lx[2] );
   glEnd();

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_y );

   glBegin(GL_LINES);
   glVertex3d( 1.01*R*ly[0], 1.01*R*ly[1], 1.01*R*ly[2] );
   glVertex3d( 1.10*R*ly[0], 1.10*R*ly[1], 1.10*R*ly[2] );
   glEnd();

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_z );

   glBegin(GL_LINES);
   glVertex3d( 1.01*R*lz[0], 1.01*R*lz[1], 1.01*R*lz[2] );
   glVertex3d( 1.10*R*lz[0], 1.10*R*lz[1], 1.10*R*lz[2] );
   glEnd();


   glEndList();
#endif
}




#ifndef TERMINAL
// GUI stuff

double OGL3dpattern::pattern_value( QString yaxis, QString pol, int idx_phi, int idx_theta ) const
{
   double dv;
   complex<double> cv;

   if ( idx_phi < 0 || idx_phi > nophi-1 || idx_theta < 0 || idx_theta > notheta-1 )
   {
      throw( "pattern_value index problem" );
   }

   int idx = idx_phi + nophi*idx_theta;

   if ( pol == tr("reference") )
      cv = RefValue[ idx ];
   else if ( pol == tr("cross") )
      cv = CrossValue[ idx ];
   else
      throw( "unknown polarization" );

   if ( yaxis == tr("pattern (dB)") )
      dv = 20.0*log10( abs(cv)/abs(MaxValue) );
   else if ( yaxis == tr("pattern (lin.)") )
      dv = abs(cv)/abs(MaxValue);
   else if ( yaxis == tr("RCS (dB)") )
      dv = 10.0*log10( 4.0*M_PI*abs(cv*conj(cv)) / (MD.wavelength*MD.wavelength) );
   else if ( yaxis == tr("RCS (lin.)") )
      dv = 4.0*M_PI*abs(cv*conj(cv)) / (MD.wavelength*MD.wavelength);
   else
      throw( "unknown r-axis" );

   dv =  R * ( dv - rmin ) / fabs( rmax - rmin );
   if ( dv < 0.0 ) dv = 0.0;

   return(dv);
}


void OGL3dpattern::false_color( double r, GLfloat *clr ) const
{
   const float alpha = 1.0;
   double v = r / R;

   if ( v < 0 - DBL_EPSILON )
   {
      clr[0] = 0.0;
      clr[1] = 0.0;
      clr[2] = 0.0;
      clr[3] = alpha;
   }
   else if ( v < 0.2 )
   {
      clr[0] = 1.0 - v/0.2;
      clr[1] = 0.0;
      clr[2] = 1.0;
      clr[3] = alpha;
   }
   else if ( v < 0.4 )
   {
      clr[0] = 0.0;
      clr[1] = (v-0.2)/0.2;
      clr[2] = 1.0;
      clr[3] = alpha;
   }
   else if ( v < 0.6 )
   {
      clr[0] = 0.0;
      clr[1] = 1.0;
      clr[2] = 1.0 - (v-0.4)/0.2;
      clr[3] = alpha;
   }
   else if ( v < 0.8 )
   {
      clr[0] = (v-0.6)/0.2;
      clr[1] = 1.0;
      clr[2] = 0.0;
      clr[3] = alpha;
   }
   else if ( v < 1.0 + DBL_EPSILON )
   {
      clr[0] = 1.0;
      clr[1] = 1.0 - (v-0.8)/0.2;
      clr[2] = 0.0;
      clr[3] = alpha;
   }
   else
   {
      clr[0] = 1.0;
      clr[1] = 1.0;
      clr[2] = 1.0;
      clr[3] = alpha;
   }

}



void OGL3dpattern::setup_ff( QString yaxis, QString pol, double cutoff )
{
   GLuint gl_list;

   makeCurrent();

// ff
   gl_list = partlist + 0;
   glNewList( gl_list, GL_COMPILE );

   GLfloat clr[4], clr2[4];
   double r, rv1[3], rv2[3], rv3[3], nv[3], theta, phi, lr[3];


   if ( yaxis == tr("pattern (dB)") )
      rmax = 0.0;
   else if ( yaxis == tr("pattern (lin.)") )
      rmax = 1.0;
   else if ( yaxis == tr("RCS (dB)") )
      rmax = 10.0*log10( 4.0*M_PI*MaxValue*MaxValue / (MD.wavelength*MD.wavelength) );
   else if ( yaxis == tr("RCS (lin.)") )
      rmax = 4.0*M_PI*MaxValue*MaxValue / (MD.wavelength*MD.wavelength);
   else
      throw( "unknown r-axis" );

   rmin = cutoff;
   if ( rmax < rmin ) rmax = rmin + DBL_EPSILON;

// cap
   glBegin(GL_TRIANGLE_FAN);

   r = pattern_value( yaxis, pol, 0, 0 );
   false_color( r, clr );

   rv1[0] = lz[0]*r;
   rv1[1] = lz[1]*r;
   rv1[2] = lz[2]*r;

   glVertex3d( rv1[0], rv1[1], rv1[2] );

   theta = M_PI / (notheta-1);
   phi = 0.0;
   r = pattern_value( yaxis, pol, 0, 1 );
   false_color( r, clr2 );

   lr[0] = sin(theta)*cos(phi) * r;
   lr[1] = sin(theta)*sin(phi) * r;
   lr[2] = cos(theta) * r;

   rv2[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
   rv2[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
   rv2[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


   phi = -2.0*M_PI / (nophi-1);
   r = pattern_value( yaxis, pol, nophi-2, 1 );

   lr[0] = sin(theta)*cos(phi) * r;
   lr[1] = sin(theta)*sin(phi) * r;
   lr[2] = cos(theta) * r;

   rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
   rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
   rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


   normal( rv1, rv2, rv3, nv );

   glNormal3d( -nv[0], -nv[1], -nv[2] );


   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );
   glVertex3d( rv1[0], rv1[1], rv1[2] );

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr2 );
   glVertex3d( rv2[0], rv2[1], rv2[2] );


   for ( int i=1; i<nophi; i++ )
   {
      phi = i*2.0*M_PI / (nophi-1);

      r = pattern_value( yaxis, pol, i, 1 );
      false_color( r, clr );
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );

      lr[0] = sin(theta)*cos(phi) * r;
      lr[1] = sin(theta)*sin(phi) * r;
      lr[2] = cos(theta) * r;

      rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
      rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
      rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];

      normal( rv1, rv2, rv3, nv );

      glNormal3d( nv[0], nv[1], nv[2] );
      glVertex3d( rv3[0], rv3[1], rv3[2] );

      rv2[0] = rv3[0];
      rv2[1] = rv3[1];
      rv2[2] = rv3[2];
   }

   glEnd();


// body
   for ( int j=1; j<notheta-2; j++ )
   {
      glBegin(GL_QUAD_STRIP);


      theta = j*M_PI / (notheta-1);
      phi = 0.0;
      r = pattern_value( yaxis, pol, 0, j );
      false_color( r, clr );

      lr[0] = sin(theta)*cos(phi) * r;
      lr[1] = sin(theta)*sin(phi) * r;
      lr[2] = cos(theta) * r;

      rv1[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
      rv1[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
      rv1[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


      theta = (j+1)*M_PI / (notheta-1);
      r = pattern_value( yaxis, pol, 0, j+1 );
      false_color( r, clr2 );

      lr[0] = sin(theta)*cos(phi) * r;
      lr[1] = sin(theta)*sin(phi) * r;
      lr[2] = cos(theta) * r;

      rv2[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
      rv2[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
      rv2[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


      theta = j*M_PI / (notheta-1);
      phi = -2.0*M_PI / (nophi-1);
      r = pattern_value( yaxis, pol, 1, j );

      lr[0] = sin(theta)*cos(phi) * r;
      lr[1] = sin(theta)*sin(phi) * r;
      lr[2] = cos(theta) * r;

      rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
      rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
      rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


      normal( rv1, rv2, rv3, nv );

      glNormal3d( -nv[0], -nv[1], -nv[2] );


      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );
      glVertex3d( rv1[0], rv1[1], rv1[2] );

      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr2 );
      glVertex3d( rv2[0], rv2[1], rv2[2] );

      for ( int i=1; i<nophi; i++ )
      {
         theta = j*M_PI / (notheta-1);
         phi = i*2.0*M_PI / (nophi-1);

         r = pattern_value( yaxis, pol, i, j );
         false_color( r, clr );
         glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );

         lr[0] = sin(theta)*cos(phi) * r;
         lr[1] = sin(theta)*sin(phi) * r;
         lr[2] = cos(theta) * r;

         rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
         rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
         rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];

         normal( rv1, rv2, rv3, nv );

         glNormal3d( nv[0], nv[1], nv[2] );
         glVertex3d( rv3[0], rv3[1], rv3[2] );

         rv1[0] = rv3[0];
         rv1[1] = rv3[1];
         rv1[2] = rv3[2];


         theta = (j+1)*M_PI / (notheta-1);

         r = pattern_value( yaxis, pol, i, j+1 );
         false_color( r, clr );
         glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );

         lr[0] = sin(theta)*cos(phi) * r;
         lr[1] = sin(theta)*sin(phi) * r;
         lr[2] = cos(theta) * r;

         rv2[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
         rv2[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
         rv2[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];

         glVertex3d( rv2[0], rv2[1], rv2[2] );
      }

      glEnd();
   }


// bottom
   glBegin(GL_TRIANGLE_FAN);

   r = pattern_value( yaxis, pol, 0, notheta-1 );
   false_color( r, clr );

   rv1[0] = -r * lz[0];
   rv1[1] = -r * lz[1];
   rv1[2] = -r * lz[2];


   theta = (notheta-2)*M_PI / (notheta-1);
   phi = -0*2.0*M_PI / (nophi-1);
   r = pattern_value( yaxis, pol, 0, notheta-2 );
   false_color( r, clr2 );

   lr[0] = sin(theta)*cos(phi) * r;
   lr[1] = sin(theta)*sin(phi) * r;
   lr[2] = cos(theta) * r;

   rv2[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
   rv2[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
   rv2[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


   phi = 2.0*M_PI / (nophi-1);
   r = pattern_value( yaxis, pol, 1, notheta-2 );

   lr[0] = sin(theta)*cos(phi) * r;
   lr[1] = sin(theta)*sin(phi) * r;
   lr[2] = cos(theta) * r;

   rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
   rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
   rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];


   normal( rv1, rv2, rv3, nv );

   glNormal3d( -nv[0], -nv[1], -nv[2] );


   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );
   glVertex3d( rv1[0], rv1[1], rv1[2] );

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr2 );
   glVertex3d( rv2[0], rv2[1], rv2[2] );


   for ( int i=1; i<nophi; i++ )
   {
      phi = -i*2.0*M_PI / (nophi-1);

      r = pattern_value( yaxis, pol, nophi-1-i, notheta-2 );
      false_color( r, clr );
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );

      lr[0] = sin(theta)*cos(phi) * r;
      lr[1] = sin(theta)*sin(phi) * r;
      lr[2] = cos(theta) * r;

      rv3[0] = lx[0]*lr[0] + ly[0]*lr[1] + lz[0]*lr[2];
      rv3[1] = lx[1]*lr[0] + ly[1]*lr[1] + lz[1]*lr[2];
      rv3[2] = lx[2]*lr[0] + ly[2]*lr[1] + lz[2]*lr[2];

      normal( rv1, rv2, rv3, nv );

      glNormal3d( nv[0], nv[1], nv[2] );
      glVertex3d( rv3[0], rv3[1], rv3[2] );

      rv2[0] = rv3[0];
      rv2[1] = rv3[1];
      rv2[2] = rv3[2];
   }

   glEnd();


   glEndList();
}


void OGL3dpattern::setup_geo_axes()
{
   GLuint gl_list;

   makeCurrent();

// geo
   gl_list = partlist + 1;
   glNewList( gl_list, GL_COMPILE );

   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_matrix() )
      {
         try { ip->matrix_add_to_OGLlist( &MD, gl_list ); }
         catch ( const char *error )
         {
            Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            break;
         }
      }

      if ( ip->use_RHS() )
      {
         try { ip->RHS_add_to_OGLlist( &MD, gl_list ); }
         catch ( const char *error )
         {
            Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            break;
         }
      }
   }

   glEndList();


// axes
   const GLfloat clr_origin[4] = { 1.00, 0.00, 1.00, 1.00 };

   gl_list = partlist + 2;
   glNewList( gl_list, GL_COMPILE );

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_origin );

   glBegin(GL_LINES);

   glNormal3d( 0.0, 0.0, 1.0 );

   glVertex3d( -0.10*R, 0.0, 0.0 );
   glVertex3d(  0.60*R, 0.0, 0.0 );

   glVertex3d( 0.0, -0.10*R, 0.0 );
   glVertex3d( 0.0,  0.60*R, 0.0 );

   glVertex3d( 0.0, 0.0, -0.10*R );
   glVertex3d( 0.0, 0.0,  1.00*R );

   glEnd();

   glEndList();


// circumsphere
   setViewRadius(R);
}


void OGL3dpattern::setup_level( double lvl )
{
   GLuint gl_list;
   const GLfloat clr[4]   = { 0.50, 0.50, 0.50, 0.60 };
   const int nophi = 30;
   const int notheta = 20;

   double theta, phi, rv[3];

   double r = R * ( lvl - rmin ) / ( rmax - rmin );

   makeCurrent();

   gl_list = partlist + 4;
   glNewList( gl_list, GL_COMPILE );

   glEnable( GL_BLEND );
   glEnable(GL_CULL_FACE);
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );
   glCullFace( GL_BACK );


   glBegin(GL_TRIANGLE_FAN);

   glNormal3d( 0.0, 0.0, r );
   glVertex3d( 0.0, 0.0, r );

   theta = M_PI / (notheta-1);
   for ( int i=0; i<nophi+1; i++ )
   {
      phi = i*2.0*M_PI / nophi;

      rv[0] = sin(theta)*cos(phi);
      rv[1] = sin(theta)*sin(phi);
      rv[2] = cos(theta);

      glNormal3d( rv[0], rv[1], rv[2] );
      glVertex3d( r*rv[0], r*rv[1], r*rv[2] );
   }

   glEnd();


   for ( int j=1; j<notheta-2; j++ )
   {
      glBegin(GL_QUAD_STRIP);

      for ( int i=0; i<nophi+1; i++ )
      {
         theta = j*M_PI / (notheta-1);
         phi = i*2.0*M_PI / nophi;

         rv[0] = sin(theta)*cos(phi);
         rv[1] = sin(theta)*sin(phi);
         rv[2] = cos(theta);

         glNormal3d( rv[0], rv[1], rv[2] );
         glVertex3d( r*rv[0], r*rv[1], r*rv[2] );


         theta = (j+1)*M_PI / (notheta-1);

         rv[0] = sin(theta)*cos(phi);
         rv[1] = sin(theta)*sin(phi);
         rv[2] = cos(theta);

         glNormal3d( rv[0], rv[1], rv[2] );
         glVertex3d( r*rv[0], r*rv[1], r*rv[2] );
      }

      glEnd();
   }


   glBegin(GL_TRIANGLE_FAN);

   glNormal3d( 0.0, 0.0, -r );
   glVertex3d( 0.0, 0.0, -r );

   theta = (notheta-2)*M_PI / (notheta-1);
   for ( int i=0; i<nophi+1; i++ )
   {
      phi = -i*2.0*M_PI / nophi;

      rv[0] = sin(theta)*cos(phi);
      rv[1] = sin(theta)*sin(phi);
      rv[2] = cos(theta);

      glNormal3d( rv[0], rv[1], rv[2] );
      glVertex3d( r*rv[0], r*rv[1], r*rv[2] );
   }

   glEnd();



   glDisable(GL_CULL_FACE);
   glDisable( GL_BLEND );
   glEndList();

}



void OGL3dpattern::display_pattern( bool ff, bool geo, bool axes, bool lvl, bool refplane )
{
   clearall();

   lvlM = lvl;

   glNewList( object, GL_COMPILE );

   if ( ff )       glCallList( partlist + 0 );
   if ( geo )      glCallList( partlist + 1 );
   if ( axes )     glCallList( partlist + 2 );
   if ( refplane ) glCallList( partlist + 3 );
   if ( lvl )      glCallList( partlist + 4 );

   glEndList();

   updateGL();
}



void OGL3dpattern::wheelEvent( QWheelEvent *e )
{
   if ( lvlM )
   {
      emit got_wheel_event( e->delta() );
      e->accept();
   }
   else
      OGLdrawing::wheelEvent(e);
}

void OGL3dpattern::mousePressEvent( QMouseEvent* e )
{
   if ( lvlM && e->button() == Qt::MidButton )
   {
      mpy = e->y();
      e->accept();
   }
   else
      OGLdrawing::mousePressEvent( e );
}

void OGL3dpattern::mouseMoveEvent( QMouseEvent* e )
{
   if ( lvlM && e->state() == Qt::MidButton )
   {
      emit got_wheel_event( mpy - e->y() );

      mpy = e->y();
      e->accept();
   }
   else
      OGLdrawing::mouseMoveEvent( e );

}



QString OGL3dpattern::GetSavePatternName()
{
   QString fn;
   QFileDialog fd( MD.PrjPath,
                   tr("All files (*)"),
                   this,
                   tr("save image dialog"),
                   true );
   fd.setMode( QFileDialog::AnyFile );
   fd.setCaption(tr("Save image"));

   QStringList list = QImage::outputFormatList();
   QStringList::Iterator it = list.begin();
   while( it != list.end() )
   {
      fd.addFilter( tr("Image (*.") + *it + " *." + (*it).lower() + ")" );
      ++it;
   }
   fd.addFilter( tr("Data (*.FFSPH *.ffsph)") );
   fd.setSelectedFilter(tr("All files (*)"));

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}


void OGL3dpattern::export_image()
{
   fstream file;
   QImage IMG = grabFrameBuffer(true);
   QString fn = GetSavePatternName();

   if ( fn != "" )
   {
      QString ext = fn.section( '.', -1 );

      if ( ext.upper() == "FFSPH" )
      {
         file.open( sysCodec->fromUnicode(fn), ios::out );
         file.precision(15);

         write_ffsph(file);

         if ( file.fail() )
         {
            Xlog << "cannot save \"" << fn << "!" << endl;
         }
         else
         {
            Xlog << "pattern data saved as \"" << fn << "\"." << endl;
         }
         file.close();
      }
      else
      {
         if ( IMG.save( fn, ext.upper(), 100 ) )
         {
            Xlog << "pattern exported as \"" << fn << "\"." << endl;
         }
         else
         {
            Xlog << "cannot save \"" << fn << "\", wrong extension?" << endl;
         }
      }
   }
}





// end GUI stuff
#endif

