//
// OpenGL Widget (3D pattern)
// (c) 29.8.2006 Ingo D. Rullhusen
//

#ifndef OGL3DPATTERN_H
#define OGL3DPATTERN_H

#include <GL/gl.h>

#include "ogldrawing.h"

#include "farfield.h"


class OGL3dpattern: public OGLdrawing
{
   Q_OBJECT

   public:
      OGL3dpattern( QWidget* parent, const char* name );
      ~OGL3dpattern();


   public slots:
      void clearall();

      void init_ff( FARFIELD *FF, QString cosys );   // (refplane, cosys)

      void setup_ff( QString yaxis, QString pol, double cutoff ); // OGL 0 (Yaxis, pol, min cutoff)
      void setup_geo_axes();   // OGL 1 2
      void setup_refplane( double theta, double phi, double psi ); // OGL 3
      void setup_level( double lvl );      // OGL 4

      void display_pattern( bool ff, bool geo, bool axes, bool lvl, bool refplane );

      void export_image();

   protected:
      void wheelEvent( QWheelEvent *e );
      void mousePressEvent( QMouseEvent* e );
      void mouseMoveEvent( QMouseEvent* e );

   signals:
      void got_wheel_event( int delta );

   private:
      GLuint partlist;
      double R;

      double lx[3], ly[3], lz[3];   // dipole/ref global x -> local z

      QString coordinates;
      int notheta, nophi;
      complex<double> *RefValue, *CrossValue;
      double MaxValue;

      bool lvlM;
      int mpy;
      double rmax, rmin;

      double pattern_value( QString yaxis, QString pol, int idx_phi, int idx_theta ) const;
      void false_color( double r, GLfloat *clr ) const;

      QString GetSavePatternName();
      void write_ffsph( fstream &file ) const;
};


#endif
