//
// Base class for OpenGL drawing
// (c) 31.5.2007 Ingo D. Rullhusen
//

#ifndef OGLDRAWING_H
#define OGLDRAWING_H

#include <qgl.h>

class OGLdrawing : public QGLWidget
{
   Q_OBJECT
   Q_PROPERTY( QColor BackgroundColor READ BkgColor WRITE setBkgColor )
   Q_PROPERTY( bool anaglyph READ Anaglyph WRITE setAnaglyph )

public:
   OGLdrawing( QWidget* parent, const char* name );
   ~OGLdrawing();

   QColor BkgColor() const;
   bool Anaglyph() const;

   double ViewRadius() const;

public slots:
   void reset_rotation();

   void setBkgColor( QColor c );
   void setAnaglyph( bool a );

   void setViewRadius( double r );

   void initializeGL();
   void paintGL();
   void resizeGL( int w, int h );

protected:
   void update_view_radius();

   void focusInEvent( QFocusEvent* e );

   void mousePressEvent( QMouseEvent* e );
   void mouseMoveEvent( QMouseEvent* e );
   void wheelEvent( QWheelEvent* e );

   GLuint object;  // to be set in derived class

signals:
   void gotFocus();

private:
   bool anaglyph;
   int w, h;
   double R0, R;
   QColor bkgC;
   QPoint initMousePos;
   GLfloat initHaxis[3], initVaxis[3], RM[16], TM[16], PML[16], PMC[16], PMR[16], mpy;
   GLfloat HRot, VRot, scale_x, scale_y, scale_z;

   GLvoid *pixels;
};

#endif
