//
// OpenGL Widget (scan)
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <vector>

#include "global_vars.h"

#include "oglscan.h"

#include "rotmat.h"


#define MAX_OGL_OBJ 2


OGLscan::OGLscan( QWidget* parent, const char* name )
   : OGLdrawing::OGLdrawing( parent, name )
{
   makeCurrent();
   partlist = glGenLists( MAX_OGL_OBJ );
}

OGLscan::~OGLscan()
{
   makeCurrent();
   glDeleteLists( partlist, MAX_OGL_OBJ );
}


void OGLscan::clearall()
{
   makeCurrent();
   glDeleteLists( object, 1 );
   object = glGenLists( 1 );
}


void OGLscan::setup_geo()
{
   double R;
   GLuint gl_list;
   
   R = 0.75*sqrt( MD.size_x*MD.size_x
		  + MD.size_y*MD.size_y
		  + MD.size_z*MD.size_z );
   
   makeCurrent();
   
// geo
   gl_list = partlist + 0;
   glNewList( gl_list, GL_COMPILE );
   
   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
   ip != used_modules.end();
   ip++ )
   {
      if ( ip->use_matrix() )
      {
	 try { ip->matrix_add_to_OGLlist( &MD, gl_list ); }
	 catch ( const char *error )
	 {
	    Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
		  << ip->name() << " ): "
		  << error << "!" << endl;
	    break;
	 }
      }
      
      if ( ip->use_RHS() )
      {
	 try { ip->RHS_add_to_OGLlist( &MD, gl_list ); }
	 catch ( const char *error )
	 {
	    Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
		  << ip->name() << " ): "
		  << error << "!" << endl;
	    break;
	 }
      }
   }
   
   glEndList();
   
   
// circumsphere
   setViewRadius(R);
   
}


void dreibein( double rmat[3][3], double R )
{
   const GLfloat clr_ref[4]   = { 1.00, 0.00, 0.00, 1.00 };
   const GLfloat clr_cross[4] = { 0.25, 0.25, 1.00, 1.00 };
   const GLfloat clr_R[4]     = { 0.90, 0.90, 0.00, 1.00 };
   
   const GLfloat clr_mark[4]  = { 0.00, 1.00, 0.00, 1.00 };
   
   double V[3];
   
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_ref );
   
   V[0] = rmat[0][0];
   V[1] = rmat[1][0];
   V[2] = rmat[2][0];
   
   glBegin(GL_LINES);
   
   glVertex3d( 1.01*R*rmat[0][2] - 0.02*R*V[0],
	       1.01*R*rmat[1][2] - 0.02*R*V[1],
	       1.01*R*rmat[2][2] - 0.02*R*V[2] );
   
   glVertex3d( 1.01*R*rmat[0][2] + 0.2*R*V[0],
	       1.01*R*rmat[1][2] + 0.2*R*V[1],
	       1.01*R*rmat[2][2] + 0.2*R*V[2] );
   
   glEnd();
   
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_cross );
   
   V[0] = rmat[0][1];
   V[1] = rmat[1][1];
   V[2] = rmat[2][1];
   
   glBegin(GL_LINES);
   
   glVertex3d( 1.01*R*rmat[0][2] - 0.02*R*V[0],
	       1.01*R*rmat[1][2] - 0.02*R*V[1],
	       1.01*R*rmat[2][2] - 0.02*R*V[2] );
   
   glVertex3d( 1.01*R*rmat[0][2] + 0.2*R*V[0],
	       1.01*R*rmat[1][2] + 0.2*R*V[1],
	       1.01*R*rmat[2][2] + 0.2*R*V[2] );
   
   glEnd();
   
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_R );
   
   V[0] = rmat[0][2];
   V[1] = rmat[1][2];
   V[2] = rmat[2][2];
   
   glBegin(GL_LINES);
   
   glVertex3d( 1.01*R*rmat[0][2] - 0.02*R*V[0],
	       1.01*R*rmat[1][2] - 0.02*R*V[1],
	       1.01*R*rmat[2][2] - 0.02*R*V[2] );
   
   glVertex3d( 1.01*R*rmat[0][2] + 0.2*R*V[0],
	       1.01*R*rmat[1][2] + 0.2*R*V[1],
	       1.01*R*rmat[2][2] + 0.2*R*V[2] );
   
   glEnd();
   
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_mark );
   
   glBegin(GL_LINES);
   
   glVertex3d( 0.3*R*V[0], 0.3*R*V[1], 0.3*R*V[2] );
   glVertex3d( 0.9*R*V[0], 0.9*R*V[1], 0.9*R*V[2] );
   
   glEnd();
}


void OGLscan::display_scan( double theta, double phi, double psi,
			    double pol,
			    double start, double stop, int samples )
{
   clearall();
   
   
// scan geo
   double sw, rmat[3][3], V[3];
   const int noscanplp = 30;
   const GLfloat clr_plane_1[4] = { 0.70, 0.70, 0.70, 0.50 };
   const GLfloat clr_plane_2[4] = { 0.40, 0.40, 0.40, 0.50 };
   
   const GLfloat clr_arrow_1[4] = { 0.70, 0.30, 0.00, 1.00 };
   const GLfloat clr_arrow_2[4] = { 0.00, 1.00, 0.00, 1.00 };
   
   const GLfloat clr_tics[4] = { 1.00, 1.00, 0.00, 1.00 };
   
   const GLfloat clr_origin[4] = { 1.00, 0.00, 1.00, 1.00 };
   
   double R = 1.1*sqrt( MD.size_x*MD.size_x
			+ MD.size_y*MD.size_y
			+ MD.size_z*MD.size_z );
   
   GLuint gl_list = partlist + 1;
   glNewList( gl_list, GL_COMPILE );
   
   
// origin
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_origin );
   
   glBegin(GL_LINES);
   
   glVertex3d( -0.10*R, 0.0, 0.0 );
   glVertex3d(  0.60*R, 0.0, 0.0 );
   
   glVertex3d( 0.0, -0.10*R, 0.0 );
   glVertex3d( 0.0,  0.60*R, 0.0 );
   
   glVertex3d( 0.0, 0.0, -0.10*R );
   glVertex3d( 0.0, 0.0,  1.00*R );
   
   glEnd();
   
   
// center
   rotmat( rmat, 1.0, pol, 0.0, psi, theta, phi );
   
   dreibein( rmat, R );
   
   rotmat( rmat, 1.0, 0.0, 0.0, psi, theta, phi );
   
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_arrow_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_2 );
   
   glBegin(GL_QUADS);
   
   glNormal3d( rmat[0][0], rmat[1][0], rmat[2][0] );
   
   V[0] = 0.75*R*rmat[0][2];
   V[1] = 0.75*R*rmat[1][2];
   V[2] = 0.75*R*rmat[2][2];
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2],
	       V[1] + 0.1*R*rmat[1][2],
	       V[2] + 0.1*R*rmat[2][2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1],
	       V[1] - 0.1*R*rmat[1][1],
	       V[2] - 0.1*R*rmat[2][1] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2],
	       V[1] - 0.1*R*rmat[1][2],
	       V[2] - 0.1*R*rmat[2][2] );
   glVertex3d( V[0] + 0.1*R*rmat[0][1],
	       V[1] + 0.1*R*rmat[1][1],
	       V[2] + 0.1*R*rmat[2][1] );
   
   glEnd();
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_1 );
   
   glBegin(GL_TRIANGLES);
   
   glNormal3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][1],
	       V[1] + 0.1*R*rmat[1][1],
	       V[2] + 0.1*R*rmat[2][1] );
   glVertex3d( V[0] + 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][0] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1],
	       V[1] - 0.1*R*rmat[1][1],
	       V[2] - 0.1*R*rmat[2][1] );
   
   glNormal3d( rmat[0][1], rmat[1][1], rmat[2][1] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2],
	       V[1] + 0.1*R*rmat[1][2],
	       V[2] + 0.1*R*rmat[2][2] );
   glVertex3d( V[0] + 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][0] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2],
	       V[1] - 0.1*R*rmat[1][2],
	       V[2] - 0.1*R*rmat[2][2] );
   
   glEnd();
   
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_2 );
   
   glBegin(GL_TRIANGLES);
   
   glNormal3d( rmat[0][1], rmat[1][1], rmat[2][1] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2] - 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][2] - 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][2] - 0.1*R*rmat[2][0] );
   glVertex3d( V[0],
	       V[1],
	       V[2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2] - 0.1*R*rmat[0][0],
	       V[1] - 0.1*R*rmat[1][2] - 0.1*R*rmat[1][0],
	       V[2] - 0.1*R*rmat[2][2] - 0.1*R*rmat[2][0] );
   
   glNormal3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][1] - 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][1] - 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][1] - 0.1*R*rmat[2][0] );
   glVertex3d( V[0],
	       V[1],
	       V[2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1] - 0.1*R*rmat[0][0],
	       V[1] - 0.1*R*rmat[1][1] - 0.1*R*rmat[1][0],
	       V[2] - 0.1*R*rmat[2][1] - 0.1*R*rmat[2][0] );
   
   glEnd();
   
   
// start
   rotmat( rmat, 1.0, pol, start, psi, theta, phi );
   
   dreibein( rmat, R );
   
   rotmat( rmat, 1.0, 0.0, start, psi, theta, phi );
   
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_arrow_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_2 );

   glBegin(GL_QUADS);
   
   glNormal3d( rmat[0][0], rmat[1][0], rmat[2][0] );
   
   V[0] = 0.75*R*rmat[0][2];
   V[1] = 0.75*R*rmat[1][2];
   V[2] = 0.75*R*rmat[2][2];
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2],
	       V[1] + 0.1*R*rmat[1][2],
	       V[2] + 0.1*R*rmat[2][2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1],
	       V[1] - 0.1*R*rmat[1][1],
	       V[2] - 0.1*R*rmat[2][1] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2],
	       V[1] - 0.1*R*rmat[1][2],
	       V[2] - 0.1*R*rmat[2][2] );
   glVertex3d( V[0] + 0.1*R*rmat[0][1],
	       V[1] + 0.1*R*rmat[1][1],
	       V[2] + 0.1*R*rmat[2][1] );
   
   glEnd();

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_1 );

   glBegin(GL_TRIANGLES);

   glNormal3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][1],
	       V[1] + 0.1*R*rmat[1][1],
	       V[2] + 0.1*R*rmat[2][1] );
   glVertex3d( V[0] + 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][0] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1],
	       V[1] - 0.1*R*rmat[1][1],
	       V[2] - 0.1*R*rmat[2][1] );
   
   glNormal3d( rmat[0][1], rmat[1][1], rmat[2][1] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2],
	       V[1] + 0.1*R*rmat[1][2],
	       V[2] + 0.1*R*rmat[2][2] );
   glVertex3d( V[0] + 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][0] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2],
	       V[1] - 0.1*R*rmat[1][2],
	       V[2] - 0.1*R*rmat[2][2] );
   
   glEnd();
   
   
// stop
   rotmat( rmat, 1.0, pol, stop, psi, theta, phi );
   
   dreibein( rmat, R );
   
   rotmat( rmat, 1.0, 0.0, stop, psi, theta, phi );

   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_arrow_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_2 );

   glBegin(GL_QUADS);
   
   glNormal3d( rmat[0][0], rmat[1][0], rmat[2][0] );
   
   V[0] = 0.75*R*rmat[0][2];
   V[1] = 0.75*R*rmat[1][2];
   V[2] = 0.75*R*rmat[2][2];
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2],
	       V[1] + 0.1*R*rmat[1][2],
	       V[2] + 0.1*R*rmat[2][2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1],
	       V[1] - 0.1*R*rmat[1][1],
	       V[2] - 0.1*R*rmat[2][1] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2],
	       V[1] - 0.1*R*rmat[1][2],
	       V[2] - 0.1*R*rmat[2][2] );
   glVertex3d( V[0] + 0.1*R*rmat[0][1],
	       V[1] + 0.1*R*rmat[1][1],
	       V[2] + 0.1*R*rmat[2][1] );
   
   glEnd();

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_arrow_2 );

   glBegin(GL_TRIANGLES);
   
   glNormal3d( rmat[0][1], rmat[1][1], rmat[2][1] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][2] - 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][2] - 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][2] - 0.1*R*rmat[2][0] );
   glVertex3d( V[0],
	       V[1],
	       V[2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][2] - 0.1*R*rmat[0][0],
	       V[1] - 0.1*R*rmat[1][2] - 0.1*R*rmat[1][0],
	       V[2] - 0.1*R*rmat[2][2] - 0.1*R*rmat[2][0] );
   
   glNormal3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   
   glVertex3d( V[0] + 0.1*R*rmat[0][1] - 0.1*R*rmat[0][0],
	       V[1] + 0.1*R*rmat[1][1] - 0.1*R*rmat[1][0],
	       V[2] + 0.1*R*rmat[2][1] - 0.1*R*rmat[2][0] );
   glVertex3d( V[0],
	       V[1],
	       V[2] );
   glVertex3d( V[0] - 0.1*R*rmat[0][1] - 0.1*R*rmat[0][0],
	       V[1] - 0.1*R*rmat[1][1] - 0.1*R*rmat[1][0],
	       V[2] - 0.1*R*rmat[2][1] - 0.1*R*rmat[2][0] );
   
   glEnd();


// tics
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_tics );

   glBegin(GL_LINES);
   
   for ( int i=0; i<samples; i++ )
   {
      sw = i*(stop-start) / (samples-1) + start;
      
      rotmat( rmat, 1.0, pol, sw, psi, theta, phi );
      
      glVertex3d( 0.95*R*rmat[0][2], 0.95*R*rmat[1][2], 0.95*R*rmat[2][2] );
      glVertex3d( 1.00*R*rmat[0][2], 1.00*R*rmat[1][2], 1.00*R*rmat[2][2] );
   }
   
   glEnd();
   
   
// plane
   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_plane_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_plane_2 );

   glBegin(GL_POLYGON);
   
   rotmat( rmat, R, 0.0, 0.0, psi, theta, phi );
   glNormal3d( rmat[0][1], rmat[1][1], rmat[2][1] );
   glVertex3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   
   for ( int i=1; i<noscanplp; i++ )
   {
      sw = i*2.0*M_PI/noscanplp;
      
      rotmat( rmat, R, 0.0, sw, psi, theta, phi );
      
      glVertex3d( rmat[0][2], rmat[1][2], rmat[2][2] );
   }
   
   glEnd();
   
   glDisable( GL_BLEND );
   
   
   glEndList();
   
   
// display
   glNewList( object, GL_COMPILE );
   
   glCallList( partlist + 0 );
   glCallList( partlist + 1 );
   
   glEndList();
   
   
   updateGL();
   
   emit parameter_changed( theta, phi, psi, pol, start, stop, samples );
}

