//
// Class for 2D plots
// (c) 12.9.2006 Ingo D. Rullhusen
//


#include <iostream>
using namespace std;


#include <math.h>
#include <float.h>
#include <stdio.h>
#include <qapplication.h>

#include "global_vars.h"

#include "plotting.h"


Plotting::Plotting( QWidget *parent, const char *name )
: Drawing( parent, name )
{
   ticslen = 5;
   FrC = Qt::black;
   Lmarg = Rmarg = Tmarg = Bmarg = 10;
   numsty = "%g";
   stdfont = QApplication::font(0);
   thickness = 1;
   x_min = y_min = -1.0;
   x_max = y_max = +1.0;
   marksize = 10;

   Llen = ( x_max - x_min ) * 0.1;
   nxtL_x = x_max - 1.1 * Llen;
   nxtL_y = y_max - 0.5 * Llen;
}

Plotting::~Plotting()
{
}

QColor Plotting::FrColor() const
{
    return(FrC);
}

void Plotting::setFrColor( QColor c )
{
    FrC = c;
}

int Plotting::TicsLen() const
{
    return(ticslen);
}

void Plotting::setTicsLen( int l )
{
    ticslen = l;
}

QString Plotting::NumSty() const
{
    return(numsty);
}

void Plotting::setNumSty( QString s )
{
    numsty = s;
}

QFont Plotting::StdFont() const
{
   return(stdfont);
}

void Plotting::setStdFont( QFont f )
{
   stdfont = f;
}

uint Plotting::Thickness() const
{
   return(thickness);
}

void Plotting::setThickness( uint t )
{
   thickness = t;
}

uint Plotting::Msize() const
{
   return(marksize);
}

void Plotting::setMsize( uint ms )
{
   marksize = ms;
}


void Plotting::frame( double x_min_, double x_max_,
                      uint no_xtics, const QString &xlabel,
                      double y_min_, double y_max_,
                      uint no_ytics, const QString &ylabel,
                      const QString &title )
{
   int width, height;
   QPen pen;
   QBrush brush;
   QRect bounding;

   clear();

   width = visibleWidth();
   height = visibleHeight();

   x_min = x_min_; x_max = x_max_;
   y_min = y_min_; y_max = y_max_;

   brush.setStyle( Qt::NoBrush );

   pen.setColor( FrC );
   pen.setWidth( thickness );
   pen.setStyle( Qt::SolidLine );

   QCanvasText *i;
   if ( title != NULL )
   {
      i = new QCanvasText(&canvas);
      i->setText( title );
      i->setFont( stdfont );
      i->setColor( FrC );
      i->setTextFlags( Qt::AlignCenter );
    
      bounding = i->boundingRect();
      i->setX( width/2 );
      i->setY( bounding.height()/2 + 1 );
      i->setZ( -1 );
    
      bounding = i->boundingRect();
      Tmarg = bounding.bottom() + thickness/2 + 3;
    
      i->show();
      i = NULL;
   }
    
   if ( ylabel != NULL )
   {
      i = new QCanvasText(&canvas);
      i->setText( ylabel );
      i->setFont( stdfont );
      i->setColor( FrC );
      i->setTextFlags( Qt::AlignCenter );

      bounding = i->boundingRect();
      i->setX( bounding.width()/2 + 2 );
      i->setY( height/2 );
      i->setZ( -1 );
    
      bounding = i->boundingRect();
      Lmarg = bounding.right() + 4;
    
      i->show();
      i = NULL;
   }
    
   if ( xlabel != NULL )
   {
      i = new QCanvasText(&canvas);
      i->setText( xlabel );
      i->setFont( stdfont );
      i->setColor( FrC );
      i->setTextFlags( Qt::AlignCenter );

      bounding = i->boundingRect();
      i->setX( width/2 );
      i->setY( height - bounding.height()/2 - 1 );
      i->setZ( -1 );
    
      bounding = i->boundingRect();
      Bmarg = bounding.height();
    
      i->show();
      i = NULL;
   }

// tics labels
   QCanvasText **xi = new QCanvasText*[no_xtics+2];
   for ( uint n=0; n<no_xtics+2; n++ )
   {
      i = xi[n] = new QCanvasText(&canvas);
      i->setText( sysLocale.toString( x_min + n*(x_max-x_min) / (no_xtics+1), 'g', 10 ) );
      i->setFont( stdfont );
      i->setColor( FrC );
      i->setTextFlags( Qt::AlignCenter );

      bounding = i->boundingRect();
      i->setX( Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1) );
      i->setY( height - Bmarg - bounding.height()/2 );
      i->setZ( -1 );
      i = NULL;
   }
   for ( uint n=0; n<no_xtics+2; n++ )
   {
      bounding = xi[n]->boundingRect();
      if ( bounding.top() < height - Bmarg + (int)thickness/2 )
         Bmarg = height - bounding.top() + thickness/2 + 1;
   }

   int wmax = 0;
   QCanvasText **yi = new QCanvasText*[no_ytics+2];
   for ( uint n=0; n<no_ytics+2; n++ )
   {
      i = yi[n] = new QCanvasText(&canvas);
      i->setText( sysLocale.toString( y_min + n*(y_max-y_min) / (no_ytics+1), 'g' ) );
      i->setFont( stdfont );
      i->setColor( FrC );
      i->setTextFlags( Qt::AlignCenter );

      bounding = i->boundingRect();
      if ( wmax < bounding.width() ) wmax = bounding.width();
      i->setY( height - Bmarg - ( n*(height-Tmarg-Bmarg) / (no_ytics+1) ) );
      i->setZ( -1 );
      i = NULL;
   }
   wmax = Lmarg + wmax/2;
   for ( uint n=0; n<no_ytics+2; n++ )
   {
      i = yi[n];
      i->setX( wmax );
      bounding = i->boundingRect();
      if ( bounding.right() > Lmarg - (int)thickness/2 ) Lmarg = bounding.right() + thickness/2 + 3;
      i->show();
   }

   for ( uint n=0; n<no_xtics+2; n++ )
   {
      i = xi[n];
      bounding = i->boundingRect();
      if ( n == 0 )
         i->setX( Lmarg + bounding.width()/2 );
      else if ( n == no_xtics+1 )
         i->setX( width - Rmarg - bounding.width()/2 );
      else
         i->setX( Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1) );

      i->show();
   }

   delete [] xi;
   delete [] yi;

// tics lines
   for ( uint n=1; n<no_xtics+1; n++ )
   {
      QCanvasLine *i = new QCanvasLine(&canvas);
      i->setPoints( Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1),
                       height-Bmarg,
                    Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1),
                       height-Bmarg-ticslen );
      i->setZ( -1 );
      i->setBrush( brush );
      i->setPen( pen );

      i->show();
      i = NULL;

      i = new QCanvasLine(&canvas);
      i->setPoints( Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1),
                       Tmarg,
                    Lmarg + n*(width-Lmarg-Rmarg) / (no_xtics+1),
                       Tmarg+ticslen );
      i->setZ( -1 );
      i->setBrush( brush );
      i->setPen( pen );

      i->show();
      i = NULL;
   }

   for ( uint n=1; n<no_ytics+1; n++ )
   {
      QCanvasLine *i = new QCanvasLine(&canvas);
      i->setPoints( Lmarg,
                       height - Bmarg - ( n*(height-Tmarg-Bmarg) / (no_ytics+1) ),
                    Lmarg+ticslen,
                       height - Bmarg - ( n*(height-Tmarg-Bmarg) / (no_ytics+1) ) );
      i->setZ( -1 );
      i->setBrush( brush );
      i->setPen( pen );

      i->show();
      i = NULL;

      i = new QCanvasLine(&canvas);
      i->setPoints( width-Rmarg,
                       height - Bmarg - ( n*(height-Tmarg-Bmarg) / (no_ytics+1) ),
                    width-Rmarg-ticslen,
                       height - Bmarg - ( n*(height-Tmarg-Bmarg) / (no_ytics+1) ) );
      i->setZ( -1 );
      i->setBrush( brush );
      i->setPen( pen );
    
      i->show();
      i = NULL;
   }

   QCanvasRectangle *j = new QCanvasRectangle(&canvas);
   j->setSize( width-Lmarg-Rmarg, height-Bmarg-Tmarg );
   j->setX( Lmarg );
   j->setY( Tmarg );
   j->setZ( -1 );
   j->setBrush( brush );
   j->setPen( pen );

   j->show();
   j = NULL;

// legend size and position
   Llen = ( x_max - x_min ) * 0.1;
   nxtL_x = x_max - 2 * ticslen*(x_max-x_min) / (width-Lmarg-Rmarg) - Llen;
   nxtL_y = y_max - ticslen*(y_max-y_min) / (height-Tmarg-Bmarg) - 0.03*(y_max-y_min);

}


void Plotting::plot_mark( PointStyle ps, double x, double y, QColor c, uint lw )
{
   int width, height, d, cpx, cpy;
   QPen pen;
   QBrush brush;
    
   width = visibleWidth() - Lmarg - Rmarg;
   height = visibleHeight() - Tmarg - Bmarg;

   brush.setStyle( Qt::NoBrush );

   pen.setColor( c );
   pen.setWidth( lw );
   pen.setStyle( Qt::SolidLine );
   pen.setCapStyle( Qt::SquareCap );
   pen.setJoinStyle( Qt::RoundJoin );

   cpx = Lmarg + (int)rint(width * (x-x_min) / (x_max-x_min));
   cpy = Tmarg + (int)rint(height * (1.0 - (y-y_min) / (y_max-y_min)));

   QCanvasLine *i;
   QCanvasRectangle *j;

   switch ( ps )
   {
      case NONE: break;
      case TIMES:
         {
            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            d = (int)rint( marksize / sqrt(8.0) );

            i->setPoints( cpx - d, cpy - d, cpx + d, cpy + d );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx - d, cpy + d, cpx + d, cpy - d );
            i->show();
            break;
         }
      case PLUS:
         {
            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            d = marksize/2;

            i->setPoints( cpx - d, cpy, cpx + d, cpy );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx, cpy + d, cpx, cpy - d );
            i->show();
            break;
         }
      case STAR:
         {
            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            d = (int)rint( marksize / sqrt(8.0) );

            i->setPoints( cpx - d, cpy - d, cpx + d, cpy + d );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx - d, cpy + d, cpx + d, cpy - d );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            d = marksize/2;

            i->setPoints( cpx - d, cpy, cpx + d, cpy );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx, cpy + d, cpx, cpy - d );
            i->show();
            break;
         }
      case BOX:
         {
            d = marksize/2;
            j = new QCanvasRectangle(cpx-d,cpy-d,2*d,2*d,&canvas);
            j->setBrush( brush );
            j->setPen( pen );
            j->setZ( 0 );

            j->show();
            break;
         }
      case DIAMOND:
         {
            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            d = (int)rint( marksize/sqrt(2.0) );

            i->setPoints( cpx - d, cpy, cpx, cpy + d );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx, cpy + d, cpx + d, cpy );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx + d, cpy, cpx, cpy - d );
            i->show();

            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( cpx, cpy - d, cpx - d, cpy );
            i->show();
            break;
         }
   }

}

void Plotting::plot( double *x, double *y, uint no, const QString &label,
                     QColor c, PenStyle linestyle,
                     PointStyle pointstyle, uint lw )
{
   int width, height;
   QPen pen;
   QBrush brush;
    
   width = visibleWidth() - Lmarg - Rmarg;
   height = visibleHeight() - Tmarg - Bmarg;

   brush.setStyle( Qt::NoBrush );

   pen.setColor( c );
   pen.setWidth( lw );
   pen.setStyle( linestyle );
   pen.setCapStyle( Qt::FlatCap );
   pen.setJoinStyle( Qt::RoundJoin );

   QCanvasLine *i;
   if ( x[0]>x_min-DBL_EPSILON && x[0]<x_max+DBL_EPSILON &&
        y[0]>y_min-DBL_EPSILON && y[0]<y_max+DBL_EPSILON    )
      plot_mark( pointstyle, x[0], y[0], c, lw );
   for ( uint n=1; n<no; n++ )
   {

      if ( x[n]>x_min-DBL_EPSILON && x[n]<x_max+DBL_EPSILON &&
           y[n]>y_min-DBL_EPSILON && y[n]<y_max+DBL_EPSILON    )
      {
         if ( x[n-1]>x_min-DBL_EPSILON && x[n-1]<x_max+DBL_EPSILON &&
              y[n-1]>y_min-DBL_EPSILON && y[n-1]<y_max+DBL_EPSILON    )
         {
            i = new QCanvasLine(&canvas);
            i->setBrush( brush );
            i->setPen( pen );
            i->setZ( 0 );

            i->setPoints( Lmarg + (int)rint(width * (x[n-1]-x_min) / (x_max-x_min)),
                          Tmarg + (int)rint(height * (1.0 - (y[n-1]-y_min) / (y_max-y_min))),
                          Lmarg + (int)rint(width * (x[n]-x_min) / (x_max-x_min)),
                          Tmarg + (int)rint(height * (1.0 - (y[n]-y_min) / (y_max-y_min)))
                        );

            i->show();
         }

         plot_mark( pointstyle, x[n], y[n], c, lw );
      }

   }

// legend
   if ( label != NULL )
   {
      int lpx, lpy, l;
      lpx = Lmarg + (int)rint(width * (nxtL_x-x_min) / (x_max-x_min));
      lpy = Tmarg + (int)rint(height * (1.0 - (nxtL_y-y_min) / (y_max-y_min)));
      l = (int)rint( width * Llen / (x_max-x_min) );

      i = new QCanvasLine(&canvas);
      i->setBrush( brush );
      i->setPen( pen );
      i->setZ( 1 );

      i->setPoints( lpx, lpy, lpx + l, lpy );
      i->show();

      plot_mark( pointstyle, nxtL_x+0.5*Llen, nxtL_y, c, lw );

      QCanvasText *t = new QCanvasText(&canvas);
      t->setText( label );
      t->setFont( stdfont );
      t->setColor( FrC );
      t->setTextFlags( Qt::AlignCenter );

      QRect bounding = t->boundingRect();
      t->setX( lpx - bounding.height()/2 - bounding.width()/2 );
      t->setY( lpy );
      t->setZ( 1 );
    
      t->show();

      nxtL_y -= 1.3 * bounding.height() * (y_max-y_min) / height;
   }

}
