//
// Class for 2D plots
// (c) 29.9.2004 Ingo D. Rullhusen
//

#ifndef PLOTTING_H
#define PLOTTING_H

#include "drawing.h"

typedef enum { NONE=0, TIMES, PLUS, STAR, BOX, DIAMOND } PointStyle;

class Plotting : public Drawing
{
   Q_OBJECT
   Q_PROPERTY( QColor FrameColor READ FrColor WRITE setFrColor )
   Q_PROPERTY( int TicsLength READ TicsLen WRITE setTicsLen )
   Q_PROPERTY( QString NumberStyle READ NumSty WRITE setNumSty )
   Q_PROPERTY( QFont StdFont READ StdFont WRITE setStdFont )
   Q_PROPERTY( uint LineThickness READ Thickness WRITE setThickness )
   Q_PROPERTY( uint MarkSize READ Msize WRITE setMsize )

public:
   Plotting( QWidget *parent = 0, const char *name = 0 );
   ~Plotting();
    
   void frame( double x_min, double x_max,
               uint no_xtics, const QString &xlabel,
               double y_min, double y_max,
               uint no_ytics, const QString &ylabel,
               const QString &title );

   void plot( double *x, double *y, uint no, const QString &label,
              QColor c, PenStyle lstyle, PointStyle pointstyle, uint width );

   QColor FrColor() const;
   int TicsLen() const;
   QString NumSty() const;
   QFont StdFont() const;
   uint Thickness() const;
   uint Msize() const;
    
private:
   int Lmarg, Rmarg, Tmarg, Bmarg;
   int ticslen;
   uint thickness;
   double x_min, x_max, y_min, y_max;
   QString numsty;
   QColor FrC;
   QFont stdfont;
   uint marksize;

   double nxtL_x, nxtL_y, Llen;

   void plot_mark( PointStyle ps, double x, double y, QColor c, uint width );
    
public slots:
   void setFrColor( QColor c );
   void setTicsLen( int l );
   void setNumSty( QString s );
   void setStdFont( QFont f );
   void setThickness( uint t );
   void setMsize( uint ms );

};

#endif
