//
// state class
// (c) 8.6.2007 Ingo D. Rullhusen
//

#include <unistd.h>
#include <float.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include <vector>
#include <iostream>

#include <qapplication.h>
#include <qprogressdialog.h>

#include "MyEvents.h"
#include "auxiliary_qt.h"
#include "auxiliary.h"
#include "graph.h"
#include "state.h"

#include "global_vars.h"


static const complex<double> cNULL(0.0,0.0);
static const complex<double> cONE(1.0,0.0); 
static const complex<double> cJ(0.0,1.0);   




State::State(){ state_ = INITIAL; }

State::~State(){}


void State::next_()
{
   int m;
   Graph *G;
   QString str;
   QProgressDialog *PW;

   switch( state_ )
   {

      case INITIAL:
      {
         int pp;
         double cpos_x, cpos_y, cpos_z,
                size_x, size_y, size_z;
         double llx, lly, llz, rux, ruy, ruz;

         Xlog << endl;
         Xlog << "starting geometry set-up..." << endl;

         llx = MD.cpos_x-0.5*MD.size_x;
         lly = MD.cpos_y-0.5*MD.size_y;
         llz = MD.cpos_z-0.5*MD.size_z;
         rux = MD.cpos_x+0.5*MD.size_x;
         ruy = MD.cpos_y+0.5*MD.size_y;
         ruz = MD.cpos_z+0.5*MD.size_z;

         pp = 0;
         m = 0;
         for ( vector<ExternalModule>::iterator ip = used_modules.begin();
               ip != used_modules.end();
               ip++ )
         {
            m++;

            if ( ip->use_matrix() )
            {
               try
               {
                  if ( ! thread_of_geo_setup(&*ip) )
                     throw( "canceled" );
               }
               catch ( const char *error )
               {
                  Xlog << "MODULE( " << m << ": "
                       << ip->name() << " ): ";
                  throw(error);
               }

               ip->matrix_bounding_box( &MD,
                                        &cpos_x, &cpos_y, &cpos_z,
                                        &size_x, &size_y, &size_z );
               llx = ( llx < cpos_x-0.5*size_x ? llx : cpos_x-0.5*size_x );
               lly = ( lly < cpos_y-0.5*size_y ? lly : cpos_y-0.5*size_y );
               llz = ( llz < cpos_z-0.5*size_z ? llz : cpos_z-0.5*size_z );
               rux = ( rux > cpos_x+0.5*size_x ? rux : cpos_x+0.5*size_x );
               ruy = ( ruy > cpos_y+0.5*size_y ? ruy : cpos_y+0.5*size_y );
               ruz = ( ruz > cpos_z+0.5*size_z ? ruz : cpos_z+0.5*size_z );

// renumber
               if ( MD.precon != 0 )
               {

               int *perm, *rperm, i, j, no, sno;
               double *cp_x, *cp_y, *cp_z, *s_x, *s_y, *s_z, D;

               str = QProgressDialog::tr("Search neighbours in %1").arg(ip->name());

               PW = new QProgressDialog( str,
                                         QProgressDialog::tr("Cancel"),
                                         2*ip->no_of_bfkt()-1,
                                         ptrMW,
                                         QProgressDialog::tr("progress"),
                                         true );
               PW->setProgress(0);

               no = ip->no_of_bfkt();

               G = new Graph( no );

               double *CPx = new double[ no ];
               double *CPy = new double[ no ];
               double *CPz = new double[ no ];
               double *R   = new double[ no ];

#pragma omp parallel for private(cp_x,cp_y,cp_z,s_x,s_y,s_z,sno)
               for ( i=0; i<no; ++i )
               {
                  double CP[3], r;

#pragma omp critical (PW_state_1)
{
                  PW->setProgress(i);
                  if ( PW->wasCanceled() )
                  {
                     delete PW;

                     delete G;
                     G = NULL;

                     delete [] CPz;
                     delete [] CPy;
                     delete [] CPx;
                     delete [] R;

                     Xlog << "MODULE( " << m << ": "
                          << ip->name() << " ): ";
                     throw( "neighbour search canceled" );
                  }
                  ptrAPP->processEvents();
}

                  ip->matrix_bfkt_bounding_box( &MD, i, &sno,
                                                &cp_x, &cp_y, &cp_z,
                                                &s_x, &s_y, &s_z );

                  umkugel( sno, cp_x, cp_y, cp_z, s_x, s_y, s_z, r, CP );

                  CPx[i] = CP[0];
                  CPy[i] = CP[1];
                  CPz[i] = CP[2];
                  R[i]   = r;

                  delete [] cp_x;
                  delete [] cp_y;
                  delete [] cp_z;
                  delete [] s_x;
                  delete [] s_y;
                  delete [] s_z;
               }

#pragma omp parallel for ordered private(D,j)
               for ( i=0; i<no; ++i )
               {
#pragma omp critical (PW_state_2)
{
                  PW->setProgress(no+i);
                  if ( PW->wasCanceled() )
                  {
                     delete PW;

                     delete G;
                     G = NULL;

                     delete [] CPz;
                     delete [] CPy;
                     delete [] CPx;
                     delete [] R;

                     Xlog << "MODULE( " << m << ": "
                          << ip->name() << " ): ";
                     throw( "neighbour search canceled" );
                  }
                  ptrAPP->processEvents();
}
                  D = MD.wavelength*MD.near_limit + R[i];

#pragma omp ordered
{
                  for ( j=i+1; j<no; ++j )
                  {
                     if ( HYPOT( CPx[j]-CPx[i], CPy[j]-CPy[i], CPz[j]-CPz[i] ) < D + R[j] )
                     {
                        G->insert_edge(i,j);
                     }
                  }
}
               }

               delete [] CPz;
               delete [] CPy;
               delete [] CPx;
               delete [] R;

               PW->cancel();
               delete PW;

               if ( ip->no_of_bfkt() > 0 ) G->reverse_Cuthill_McKee();

               perm = new int[ ip->no_of_bfkt() ];
               rperm = new int[ ip->no_of_bfkt() ];
               for ( i=0; i<ip->no_of_bfkt(); i++ )
               {
                  if ( G->node[i].index < 0 || G->node[i].index >= ip->no_of_bfkt() )
                     throw( "renumbering problem" );

                  perm[ i ] = G->node[i].index;
                  rperm[ G->node[i].index ] = i;
               }

               ip->matrix_renumber( &MD, perm, rperm );

               for ( i=0; i<ip->no_of_bfkt(); i++ )
               {
                  j = G->node[i].index;

                  for ( vector<Node*>::iterator nn=G->node[i].neighlist.begin();
                        nn != G->node[i].neighlist.end();
                        nn++ )
                  {
                     if ( abs(j-(*nn)->index) > pp ) pp = abs(j-(*nn)->index);
                  }
               }

               Xlog << "MODULE( " << m << ": "
                    << ip->name() << " ): actual precon profile: " << pp + 1
                    << endl;

               delete [] rperm;
               delete [] perm;
               delete G;

               }  //if ( MD.precon != 0 )

            }

            Xlog << "MODULE( " << m << ": "
                 << ip->name() << " ): no. of unknowns: " << ip->no_of_bfkt()
                 << endl;
         }

         MD.precon = ( MD.precon < 0 ? pp + 1 : MD.precon );


         MD.size_x  = 2.0*( MD.cpos_x - llx > rux - MD.cpos_x ? MD.cpos_x - llx
                                                              : rux - MD.cpos_x );
         MD.size_y  = 2.0*( MD.cpos_y - lly > ruy - MD.cpos_y ? MD.cpos_y - lly
                                                              : ruy - MD.cpos_y );
         MD.size_z  = 2.0*( MD.cpos_z - llz > ruz - MD.cpos_z ? MD.cpos_z - llz
                                                              : ruz - MD.cpos_z );
         MD.delta_x = ( MD.AIMnox > 1 ? MD.size_x / ( MD.AIMnox - 1 ) : 0.0 );
         MD.delta_y = ( MD.AIMnoy > 1 ? MD.size_y / ( MD.AIMnoy - 1 ) : 0.0 );
         MD.delta_z = ( MD.AIMnoz > 1 ? MD.size_z / ( MD.AIMnoz - 1 ) : 0.0 );


         m = 0;
         for ( int i=0; i<(int)used_modules.size()-1; i++ )
         {
            used_modules[i+1].matrix_offset() = used_modules[i].matrix_offset()
                                                + used_modules[i].no_of_bfkt();

            m = used_modules[i+1].matrix_offset() + used_modules[i+1].no_of_bfkt();
         }

         m = ( used_modules.size() > 0
               ? used_modules[ used_modules.size()-1 ].matrix_offset()
                 + used_modules[ used_modules.size()-1 ].no_of_bfkt()
               : 0 );

         Xlog << "total no. of unknowns: " << m << endl;

         QCustomEvent *e = new QCustomEvent(UPDATE_FSETTINGS_EVENT);
         QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent


         try
         {
            grf.setup( &MD, HYPOT( 0.25*MD.delta_x, 0.25*MD.delta_y, 0.25*MD.delta_z ) );
         }
         catch ( ... )
         {
            grf.clear();
            throw("trouble in Green's function");
         }


         Xlog << endl;
         Xlog << "geometry ready." << endl;

         state_ = GEO_RDY;
         break;
      }

      case GEO_RDY:
      {
         Xlog << endl;
         Xlog << "starting AIM set-up..." << endl;

         AD.clear();

         int i, no;

         m = 0;
         for ( vector<ExternalModule>::iterator ip = used_modules.begin();
               ip != used_modules.end();
               ip++ )
         {
            m++;
            str.setNum(m);
            str = QProgressDialog::tr("Calculate AIMdata of module no. %1").arg(str);

            PW = new QProgressDialog( str,
                                      QProgressDialog::tr("Cancel"),
                                      ip->no_of_bfkt()-1,
                                      ptrMW,
                                      QProgressDialog::tr("progress"),
                                      true );
            PW->setProgress(0);

            no = ip->no_of_bfkt();
#pragma omp parallel for ordered
            for ( i=0; i<no; i++ )
            {
               AIMData adt;

               try
               {
                  adt = AIMData( MD, &*ip, i );
               }
               catch ( const char *str )
               {
                  PW->cancel();
                  delete PW;

                  throw(str);
               }
               ptrAPP->processEvents();

#pragma omp ordered
               AD.push_back( adt );

#pragma omp critical (PW_state_3)
{
               PW->setProgress( i );
               if ( PW->wasCanceled() )
               {
                  delete PW;

                  AD.clear();
                  throw( "AIM calculation canceled" );
               }
}
            }

            PW->cancel();
            delete PW;
         }

         Xlog << endl;
         Xlog << "AIM ready." << endl;

         state_ = AIM_RDY;
         break;
      }

      case AIM_RDY:
      {
         Xlog << endl;
         Xlog << "starting matrix set-up..." << endl;

         try
         {
            MAT.setup();
         }
         catch( const char *str )
         {
            MAT.clear();
            throw(str);
         }

         Xlog << endl;
         Xlog << "matrix ready." << endl;

         state_ = MAT_RDY;
         break;
      }

      case MAT_RDY:
      {
         Xlog << endl;
         Xlog << "starting right hand side set-up..." << endl;

         try
         {
            SOL.setup();
         }
         catch( const char *str )
         {
            SOL.clear();
            Xlog << "RHS calculation canceled." << endl;
            throw(str);
         }

         Xlog << endl;
         Xlog << "RHS ready." << endl;

         state_ = RHS_RDY;
         break;
      }

      case RHS_RDY:
      {
         Xlog << endl;
         Xlog << "start solving..." << endl;

         int itno = MD.max_itno;
         double eps = MD.epsilon;

         try
         {
            SOL.solve( eps, itno );

            Xlog << " no. of iterations: " << itno << endl;
            Xlog << " residuum: " << eps << endl;
         }
         catch( const char *str )
         {
            Xlog << " no. of iterations: " << itno << endl;
            Xlog << " residuum: " << eps << endl;

            Xlog << "solving canceled." << endl;
            throw(str);
         }

         Xlog << endl;
         Xlog << "solved." << endl;

         state_ = SOLVED;
         break;
      }

      case SOLVED:
      {
         Xlog << endl;
         Xlog << "start extracting..." << endl;

         try
         {
            SOL.extract();
         }
         catch( const char *str )
         {
            Xlog << "extracting canceled." << endl;
            throw(str);
         }

         Xlog << endl;
         Xlog << "extracted." << endl;

         state_ = EXTRACTED;
         break;
      }

      case EXTRACTED: { break; }

      default: { throw( "invalid state" ); }
   }




}


void State::prev_()
{

   switch(state_)
   {
      case INITIAL: { break; }
      case GEO_RDY:
      {
         grf.clear();

         for ( vector<ExternalModule>::iterator ip = used_modules.begin();
               ip != used_modules.end();
               ip++ ) ip->cleanup( &MD );

         Xlog << "back to \"initial\"." << endl;
         state_ = INITIAL;
         break;
      }
      case AIM_RDY:
      {
         AD.clear();

         Xlog << "back to \"geometry ready\"." << endl;
         state_ = GEO_RDY;
         break;
      }
      case MAT_RDY:
      {
         MAT.clear();

         Xlog << "back to \"AIM calculated\"." << endl;
         state_ = AIM_RDY;
         break;
      }
      case RHS_RDY:
      {
         SOL.clear();

         Xlog << "back to \"matrices calculated\"." << endl;
         state_ = MAT_RDY;
         break;
      }
      case SOLVED:
      {
         Xlog << "back to \"RHS ready\"." << endl;
         state_ = RHS_RDY;
         break;
      }
      case EXTRACTED:
      {
         Xlog << "back to \"solved\"." << endl;


         state_ = SOLVED;
         break;
      }
      default: { throw( "invalid state" ); }
   }




}



void State::next()
{
   next_();

   QCustomEvent *e = new QCustomEvent(UPDATE_STATE_EVENT);
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent
}

void State::prev()
{
   prev_();

   QCustomEvent *e = new QCustomEvent(UPDATE_STATE_EVENT);
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent
}

void State::back_to( StateValue s )
{
   while( state_ > s ){ prev_(); }

   QCustomEvent *e = new QCustomEvent(UPDATE_STATE_EVENT);
   QApplication::postEvent( ptrMW, e );   // *e deleted by postEvent
}



void State::write( fstream &file ) const
{
   fstream tmpf;

   file << "[State]" << endl;
   file << state_ << endl;

// tmp file states
   switch( state_ )
   {
      case EXTRACTED:

      case SOLVED:
      case RHS_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".SOL.bin";

         tmpf.open( sysCodec->fromUnicode(name), ios::out );
         if ( ! tmpf.is_open() ) throw( "cannot open SOL tmp file" );

         SOL.write_binary( tmpf );

         tmpf.close();

         file << checksum( name ) << endl;
      }

      case MAT_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".MAT.bin";

         tmpf.open( sysCodec->fromUnicode(name), ios::out );
         if ( ! tmpf.is_open() ) throw( "cannot open MAT tmp file" );

         MAT.write_binary( tmpf );

         tmpf.close();

         file << checksum( name ) << endl;
      }

      case AIM_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".AIM.bin";

         tmpf.open( sysCodec->fromUnicode(name), ios::out );
         if ( ! tmpf.is_open() ) throw( "cannot open AIM tmp file" );

         int no = AD.size();
         tmpf.write( (char*)&no, sizeof(no) );

         for ( int i=0; i<no; i++ )
         {
            AD[i].write_binary( tmpf );
         }

         tmpf.close();

         file << checksum( name ) << endl;
      }

      case GEO_RDY:
      case INITIAL:
      default: {}
   }

}

void State::read( fstream &file )
{
   fstream tmpf;

   string sbuff = "";
   while ( file.good() && sbuff != "State" )
   { sbuff = get_section_label( file, true ); }

   if ( sbuff != "State" ) throw( "cannot find State entry" );

   int s_t;
   file >> s_t;

   back_to( INITIAL );
// tmp file states
   switch( s_t )
   {
      case EXTRACTED:

      case SOLVED:
      case RHS_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".SOL.bin";
         file >> sbuff;

         SOL.clear();
         try
         {
            if ( sbuff != checksum( name ) ) throw( "checksum mismatch" );

            tmpf.open( sysCodec->fromUnicode(name), ios::in );

            if ( ! tmpf.is_open() ) throw( "cannot open SOL tmp file" );

            SOL.read_binary( tmpf );

            if ( tmpf.bad() ) throw( "SOL tmp file corrupt" );
         }
         catch ( const char *str )
         {
            Xlog << "load error: " << str << "!" << endl;
            Xlog << "back to \"matrices calculated\"" << endl;
            s_t = MAT_RDY;
         }

         if ( tmpf.is_open() ) tmpf.close();
      }

      case MAT_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".MAT.bin";
         file >> sbuff;

         try
         {
            if ( sbuff != checksum( name ) ) throw( "checksum mismatch" );

            tmpf.open( sysCodec->fromUnicode(name), ios::in );

            if ( ! tmpf.is_open() ) throw( "cannot open matrix tmp file" );

            MAT.read_binary( tmpf );

            if ( tmpf.bad() ) throw( "matrix tmp file corrupt" );
         }
         catch ( const char *str )
         {
            Xlog << "load error: " << str << "!" << endl;
            Xlog << "back to \"AIM calculated\"" << endl;
            s_t = AIM_RDY;
         }

         if ( tmpf.is_open() ) tmpf.close();
      }

      case AIM_RDY:
      {
         QString name = prjfile.section( '/', -1 );
         name = MD.PrjTmpPath + "/" + name.section( '.', 0, 0 ) + ".AIM.bin";
         file >> sbuff;

         AD.clear();
         try
         {
            if ( sbuff != checksum( name ) ) throw( "checksum mismatch" );

            tmpf.open( sysCodec->fromUnicode(name), ios::in );

            if ( ! tmpf.is_open() ) throw( "cannot open AIM tmp file" );

            int no;
            tmpf.read( (char*)&no, sizeof(no) );

            for ( int i=0; i<no; i++ )
            {
               AIMData *adt = new AIMData( tmpf );
               AD.push_back( *adt );
               delete adt;
            }

            if ( tmpf.bad() ) throw( "AIM tmp file corrupt" );
         }
         catch ( const char *str )
         {
            Xlog << "load error: " << str << "!" << endl;
            Xlog << "back to \"geometry ready\"" << endl;
            s_t = GEO_RDY;
         }

         if ( tmpf.is_open() ) tmpf.close();
      }

      case GEO_RDY:
      {
         next();
         if ( s_t == EXTRACTED )
         {
            state_ = SOLVED;
            next();
         }
      }

      case INITIAL: { break; }

      default: { throw( "unknown state" ); }
   }

   state_ = (StateValue)s_t;
}


