//
// main() of gANTlab terminal version
// (c) 7.6.2007 Ingo D. Rullhusen
//


#include <stdlib.h>
#include <unistd.h>
#include <iostream>

using namespace std;


#include "qapplication.h"
#include "qtextcodec.h"


#define DEFINE_GLOBAL_VARS
#include "global_vars.h"

#include "pattern2d_.h"
Pattern2D pattern2d;

#include "pattern3d_.h"
Pattern3D pattern3d;

#include "auxiliary.h"
#include "auxiliary_qt.h"
#include "mainform.ui.h"


int main( int argc, char ** argv )
{
   int retval = 0;

   bool reset = true;
   bool prev = false;
   bool next = false;
   bool rest = true;
   bool save = false;

   bool p2d = false;
   bool p3d = false;

   bool error = false;

   std::set_new_handler(gantlab_new_error_handler);

   ptrAPP = new QApplication();
   sysCodec = new QTextCodec();
   ptrMW = NULL;

   cout << endl;
   cout << "gANTlab (terminal version " << VERSION << ")" << endl;
   cout << "GNU antenna laboratory" << endl;
   cout << "(c) 2006/7 Ingo D. Rullhusen <d01c@uni-bremen.de>" << endl;
   cout << endl;


   qt_path = getenv("QTDIR");
   root_path = getenv("GANTLAB_ROOT_DIR");

   if ( root_path == "" )
   {
      cerr << "GANTLAB_ROOT_DIR not set!" << endl;
      cerr << endl;
   }


   if ( argc == 2 )
   {
      prjfile = argv[1];
   }
   else if ( argc > 2 )
   {
      int c;
      while( (c = getopt(argc, argv, ":arnpsP:S:")) > -1 )
      {
         switch(c)
         {
            case 'a': { /* default */; break; }
            case 'r': { reset = false;
                        prev = false;
                        next = false;
                        rest = true; 
                        break;
                      }
            case 'n': { reset = false;
                        prev = false;
                        next = true;
                        rest = false; 
                        break;
                      }
            case 'p': { reset = false;
                        prev = true;
                        next = false;
                        rest = false; 
                        break;
                      }
            case 's': { save = true; 
                        break;
                      }
            case 'P': { p2d = true;
                        pattern2d.fn = optarg;
                        break;
                      }
            case 'S': { p3d = true;
                        pattern3d.fn = optarg;
                        break;
                      }
            default: { error = true; break; }
         }
      }

      if ( optind < argc )
         prjfile = argv[optind];
      else
         error = true;
   }
   else
      error = true;

   if ( error )
   {
      cerr << "call" << endl;
      cerr << endl;
      cerr << " gANTlab_terminal [-a|-r|-n|-p] [-s] [-P <data file>] [-S <FFSPH file>] <prj file>" << endl;
      cerr << endl;
      cerr << " -a \t run all" << endl;
      cerr << " -r \t run rest" << endl;
      cerr << " -n \t next" << endl;
      cerr << " -p \t previous" << endl;
      cerr << endl;
      cerr << " -s \t save state" << endl;
      cerr << endl;
      cerr << " -P <data file> \t output 2D pattern" << endl;
      cerr << " -S <FFSPH file> \t output 3D pattern" << endl;
      cerr << endl;

      return(-2);
   }


   try
   {
      mainform::load_prj_file(prjfile);

      cout << endl;
      cout << " file: " << prjfile << endl;
      cout << endl;
      cout << " frequency = " << MD.freq << " Hz" << endl;
      cout << " wavelength= " << MD.wavelength << " m" << endl;
      cout << endl;
      cout << " AIM grid  : " << MD.AIMnox << " * "
                              << MD.AIMnoy << " * "
                              << MD.AIMnoz << endl;
      cout << " near limit: " << MD.near_limit << " wavelength" << endl;
      cout << endl;

      cout << " loaded modules (no./name/version/sel.):" << endl;
      for ( vector<ExternalModule>::iterator im = used_modules.begin();
            im != used_modules.end();
            ++im )
      {
         cout << " " << ( im-used_modules.begin() + 1 ) << " / "
              << im->name() << " / "
              << im->version() << " / ";

         if ( im->use_matrix() )
            cout << "M";
         else
            cout << ".";
         if ( im->use_RHS() )
            cout << "R";
         else
            cout << ".";
         if ( im->use_extract() )
            cout << "E";
         else
            cout << ".";

         cout << endl;
      }

      if ( reset ) S.back_to(INITIAL);
      if ( prev )  S.prev();
      if ( next )  S.next();

      while( rest && S.value() != EXTRACTED ){ S.next(); }

      if ( p2d && S.value() == EXTRACTED )
      {
         pattern2d.update_scan_parameter(true);
         pattern2d.update_calc();
         pattern2d.export_image();
      }

      if ( p3d && S.value() == EXTRACTED )
      {
         pattern3d.Pattern.setup_refplane( pattern3d.theta, pattern3d.phi, pattern3d.psi );
         pattern3d.Pattern.init_ff( SOL.ff, "spherical" );
         fstream file( sysCodec->fromUnicode(pattern3d.fn), ios::out );
         file.precision(15);
         pattern3d.Pattern.write_ffsph(file);
         if ( file.fail() ) throw("cannot write FFSPH file");
      }

      if ( save ) mainform::fileSave();
   }
   catch ( const char* error )
   {
      cerr << "ERROR: " << error << "!" << endl;
      retval = -1;
   }

   fftw_cleanup();

   delete sysCodec;
   delete ptrAPP;
    
   return(retval);
}
